#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri May 18 13:22:31 2018

@author: max
"""
from __future__ import division
import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import griddata, UnivariateSpline
from math import pi
from collections import namedtuple
from shapely.geometry import LineString, Point, MultiPoint
from matplotlib.path import Path
from Functions.FindXPtMagAxis import find_xpt_mag_axis
from Functions.CalcPsiNorm import calc_psi_norm
from Functions.CalcRho2PsiInterp import calc_rho2psi_interp
from Functions.CalcGrad import calc_grad
from Functions.CalcFSA import calc_fsa
from Functions.CalcSV import calc_svion_st, calc_svel_st, calc_svrec_st, calc_svcx_st, calc_svfus
from Functions.CalcFsPerimInt import calc_fs_perim_int
from Functions.CreateSurfAreaInterp import create_surf_area_interp
from Functions.CalcTheta1D import calc_theta1d
from Functions.CalcPtsLines import calc_pts_lines
from Functions.CalcRZ import calc_RZ
from Functions.CalcKappaElong import calc_kappa_elong
from Functions.CreateVolInterp import create_vol_interp
from Functions.CalcChiJet import calc_chi_jet
from scipy.interpolate import interp1d
import GT3.constants as constants

MARKERSIZE = constants.MARKERSIZE

e = constants.elementary_charge
u_0 = constants.mu_0
m_d = 3.343583719e-27
m_t = 5.006e-27
m_c = 1.9926467e-26
m_a = 6.643e-27


# def calc_imp_rad(nz, ne, rho, Lz, calc_dVdrho, rho_start=0, rho_end=1):
#     nz_fit = UnivariateSpline(rho[:,0], nz[:,0], k=3, s=0)
#     ne_fit = UnivariateSpline(rho[:,0], ne[:,0], k=3, s=0)
#
#     return


# noinspection SpellCheckingInspection
class Core:

    sep_val_overridden = False  # type: bool
    """Has the separatrix value been overridden?"""
    sep_val = 1.0  # type: float
    """The value of the separatrix. If this is set in the input file as <1.0, the X-point calculations are omitted"""
    r2vol = None  # type: UnivariateSpline
    """The V(r) interpolator"""
    rho2vol = None  # type: UnivariateSpline
    """The V(rho) interpolator"""
    psinorm2vol = None  # type: UnivariateSpline
    """The V(psi) interpolator on [0., 1.]"""

    def __init__(self, inp):
        # type: (GT3.ReadInfile) -> GT3.Core
        # Hold the input as a class attribute for use in functions

        self.inp = inp
        self.wall_line = inp.wall_line
        self.sep_val = inp.sep_val
        if abs(1.0 - float(self.sep_val)) > 0.0001:
            self.sep_val_overridden = True
            print "The separatrix value is being overwritten. GT3.SOL will be omitted."
        # this assumes that psi is given as a square array
        psi_shape = int(np.sqrt(inp.psirz_exp[:, 0].shape[0]))  # type: int

        raw_psi_R = inp.psirz_exp[:, 0].reshape(-1, psi_shape)  # type: np.ndarray
        raw_psi_Z = inp.psirz_exp[:, 1].reshape(-1, psi_shape)  # type: np.ndarray
        try:
            raw_psi = inp.psirz_exp[:, 2].reshape(-1, psi_shape) * inp.psi_scale  # type: np.ndarray
        except AttributeError:
            raw_psi = inp.psirz_exp[:, 2].reshape(-1, psi_shape)  # type: np.ndarray

        xpt_l, xpt_u, mag_axis = find_xpt_mag_axis(self, raw_psi_R, raw_psi_Z, raw_psi)
        xpt = [xpt_l, xpt_u]
        raw_psi_norm = calc_psi_norm(raw_psi_R, raw_psi_Z, raw_psi, xpt, mag_axis)

        raw_dpsidR = np.abs(np.gradient(raw_psi_norm, raw_psi_R[0, :], axis=1))
        raw_1_R_dpsidR = raw_dpsidR / raw_psi_R
        raw_d_dR_1_R_dpsidR = np.abs(np.gradient(raw_1_R_dpsidR, raw_psi_R[0, :], axis=1))
        raw_dpsidZ = np.abs(np.gradient(raw_psi_norm, raw_psi_Z[:, 0], axis=0))
        raw_d2psidZ2 = np.abs(np.gradient(raw_dpsidZ, raw_psi_Z[:, 0], axis=0))
        raw_dpsidr = raw_dpsidR + raw_dpsidZ
        raw_j = -(raw_d_dR_1_R_dpsidR * raw_psi_R + raw_d2psidZ2) / (raw_psi_R * u_0)

        PsiData = namedtuple('PsiData', 'R Z psi psi_norm dpsidR dpsidZ dpsidr j')
        """NamedTuple containing psi data, including R, Z, psi, psi_norm, dpsidR, dpsidZ, dpsidr, j"""

        self.psi_data = PsiData(
            raw_psi_R,
            raw_psi_Z,
            raw_psi,
            raw_psi_norm,
            raw_dpsidR,
            raw_dpsidZ,
            raw_dpsidr,
            raw_j
        )

        # calculate some important points and lines
        self.pts, self.lines = calc_pts_lines(self.psi_data, xpt, self.wall_line, mag_axis, self.sep_val, core=self)

        # create interpolation functions to convert rho to psi and vice versa
        interp_fns = calc_rho2psi_interp(self.pts, self.psi_data, self.sep_val)

        self.rho2psi = interp_fns[0]  # type: interp1d
        self.rho2psinorm = interp_fns[1]  # type: interp1d
        self.psi2rho = interp_fns[2]  # type: interp1d
        self.psi2psinorm = interp_fns[3]  # type: interp1d
        self.psinorm2rho = interp_fns[4]  # type: interp1d
        self.psinorm2psi = interp_fns[5]  # type: interp1d

        self.R0_a = self.pts.axis.mag[0]
        self.R0_g = self.pts.axis.geo[0]


        # specify rho values
        try:
            rho1d = np.concatenate((np.linspace(0, inp.edge_rho, inp.rhopts_core, endpoint=False),
                                    np.linspace(inp.edge_rho, self.sep_val, inp.rhopts_edge)), axis=0)
        except AttributeError:
            try:
                rho1d = np.linspace(0, self.sep_val, inp.rhopts)
            except AttributeError:
                raise AttributeError("You haven't specified the number of radial points.")
        self.rhopts = len(rho1d)

        theta1d, theta_markers = calc_theta1d(self.pts, inp.thetapts_approx)
        theta_xpt = theta_markers[3]
        self.thetapts = len(theta1d)

        # create rho and theta arrays (initializing the main computational grid)
        self.theta, self.rho = np.meshgrid(theta1d, rho1d)


        self.psi = self.rho2psi(self.rho)
        self.psi_norm = self.rho2psinorm(self.rho)

        # estimate elongation and triangularity
        self.kappa_vals, self.tri_vals = calc_kappa_elong(self.psi_data, np.asarray(self.lines.sep_closed.coords))

        # Calculate plasma radius considering potential for wall limiter
        # Inboard limiter
        if self.pts.ibmp[0] < self.inp.wall_line.bounds[0]:
            self.a = np.average((self.pts.obmp[0] - self.pts.axis.mag[0], self.pts.axis.mag[0] - self.inp.wall_line.bounds[0]))
        # Outboard limiter
        elif self.pts.obmp[0] >= self.inp.wall_line.bounds[2]:
            self.a = np.average((self.inp.wall_line.bounds[2] - self.pts.axis.mag[0], self.pts.axis.mag[0] - self.pts.ibmp[0]))
        else:
            self.a = np.average((self.pts.obmp[0] - self.pts.axis.mag[0], self.pts.axis.mag[0] - self.pts.ibmp[0]))

        self.shaf_shift = (self.pts.axis.mag[0] - self.pts.axis.geo[0]) / self.a
        self.r = self.rho * self.a

        self.R, self.Z = calc_RZ(self, self.rho, self.theta, theta_xpt, self.pts, self.psi_data, self.psi_norm, self.lines)

        self.xpt_loc = np.where(self.Z == np.amin(self.Z))
        self.obmp_loc = np.where(self.R == np.amax(self.R))
        self.ibmp_loc = np.where(self.R == np.amin(self.R))
        self.top_loc = np.where(self.Z == np.amax(self.Z))



        # create interpolation functions to obtain the flux surface surface area for any value of r, rho, or psi_norm
        self.r2sa, self.rho2sa, self.psinorm2sa = create_surf_area_interp(self.rho2psinorm,
                                                                          self.psinorm2rho,
                                                                          self.psi_data,
                                                                          np.asarray(self.lines.sep_closed.coords),
                                                                          self.R0_a,
                                                                          self.a,
                                                                          self.sep_val)

        # create interpolation functions to obtain the plasma volume for any value of r, rho, or psi_norm
        self.r2vol, self.rho2vol, self.psinorm2vol = create_vol_interp(self.rho2psinorm,
                                                                       self.psinorm2rho,
                                                                       self.psi_data,
                                                                       np.asarray(self.lines.sep_closed.coords),
                                                                       self.R0_a,
                                                                       self.a,
                                                                       self.sep_val)

        np.savetxt('psi2rho.txt', np.column_stack((np.linspace(0, self.sep_val, 100),
                                                   self.psi2rho(np.linspace(0, self.sep_val, 100)))))

        # initialize volume as a function of rho
        self.vol_rho = self.rho2vol(self.rho)
        """The volume as a function of rho V(rho)"""

        self.dVdrho = UnivariateSpline(np.linspace(0, self.sep_val, 100), self.rho2vol(np.linspace(0, self.sep_val, 100)), k=3,
                                       s=0).derivative()
        """The dV/drho(rho) interpolator interpolated on [0.,1.]"""

        self.dVdr = UnivariateSpline(np.linspace(0, self.sep_val, 100), self.dVdrho(np.linspace(0, self.sep_val, 100)) / self.a, k=3, s=0)
        """The dV/r(rho) interpolator interpolated on [0.,1.]"""

        self.vol = self.rho2vol(self.sep_val)
        """THe total volume of the plasma"""

        # initialize ionization rate arrays with zero
        self.izn_rate = namedtuple('izn_rate', 's t tot')(
            np.zeros(self.rho.shape),  # slow
            np.zeros(self.rho.shape),  # thermal
            np.zeros(self.rho.shape)  # total
        )

        # initialize cooling rate array with zero
        self.cool_rate = np.zeros(self.rho.shape)

        # initialize main density object

        # deuterium density
        try:
            nD = UnivariateSpline(inp.nD_data[:, 0], inp.nD_data[:, 1], k=3, s=0)(self.rho)
        except (AttributeError, TypeError):
            nD = np.zeros(self.rho.shape)

        # tritium density
        try:
            nT = UnivariateSpline(inp.nT_data[:, 0], inp.nT_data[:, 1], k=3, s=0)(self.rho)
        except (AttributeError, TypeError):
            nT = np.zeros(self.rho.shape)

        ni = nD + nT

        # electron density
        try:
            ne = UnivariateSpline(inp.ne_data[:, 0], inp.ne_data[:, 1], k=3, s=0)(self.rho)
        except (AttributeError, TypeError):
            ne = np.zeros(self.rho.shape)

        # carbon density
        try:
            nC = UnivariateSpline(inp.nC_data[:, 0], inp.nC_data[:, 1], k=3, s=0)(self.rho)
        except (AttributeError, TypeError):
            try:
                frac_C = UnivariateSpline(inp.frac_C_data[:, 0], inp.frac_C_data[:, 1], k=3, s=0)(self.rho)
                nC = ne * frac_C
            except (AttributeError, TypeError):
                nC = np.zeros(self.rho.shape)

        # tungsten density
        try:
            nW = UnivariateSpline(inp.nW_data[:, 0], inp.nW_data[:, 1], k=3, s=0)(self.rho)
        except (AttributeError, TypeError):
            try:
                frac_W = UnivariateSpline(inp.frac_W_data[:, 0], inp.frac_W_data[:, 1], k=3, s=0)(self.rho)
                nW = ne * frac_W
            except (AttributeError, TypeError):
                nW = np.zeros(self.rho.shape)

        # beryllium density
        try:
            nBe = UnivariateSpline(inp.nBe_data[:, 0], inp.nBe_data[:, 1], k=3, s=0)(self.rho)
        except (AttributeError, TypeError):
            try:
                frac_Be = UnivariateSpline(inp.frac_Be_data[:, 0], inp.frac_Be_data[:, 1], k=3, s=0)(self.rho)
                nBe = ne * frac_Be
            except (AttributeError, TypeError):
                nBe = np.zeros(self.rho.shape)

        # neon density
        try:
            nNe = UnivariateSpline(inp.nNe_data[:, 0], inp.nNe_data[:, 1], k=3, s=0)(self.rho)
        except (AttributeError, TypeError):
            try:
                frac_Ne = UnivariateSpline(inp.frac_Ne_data[:, 0], inp.frac_Ne_data[:, 1], k=3, s=0)(self.rho)
                nNe = ne * frac_Ne
            except (AttributeError, TypeError):
                nNe = np.zeros(self.rho.shape)

        # krypton density
        try:
            nKr = UnivariateSpline(inp.nKr_data[:, 0], inp.nKr_data[:, 1], k=3, s=0)(self.rho)
        except (AttributeError, TypeError):
            try:
                frac_Kr = UnivariateSpline(inp.frac_Kr_data[:, 0], inp.frac_Kr_data[:, 1], k=3, s=0)(self.rho)
                nKr = ne * frac_Kr
            except (AttributeError, TypeError):
                nKr = np.zeros(self.rho.shape)

        # argon density
        try:
            nAr = UnivariateSpline(inp.nAr_data[:, 0], inp.nAr_data[:, 1], k=3, s=0)(self.rho)
        except (AttributeError, TypeError):
            try:
                frac_Ar = UnivariateSpline(inp.frac_Ar_data[:, 0], inp.frac_Ar_data[:, 1], k=3, s=0)(self.rho)
                nAr = ne * frac_Ar
            except (AttributeError, TypeError):
                nAr = np.zeros(self.rho.shape)

        # alpha density
        try:
            na = UnivariateSpline(inp.na_data[:, 0], inp.na_data[:, 1], k=3, s=0)(self.rho)
        except (AttributeError, TypeError):
            try:
                frac_a = UnivariateSpline(inp.frac_a_data[:, 0], inp.frac_a_data[:, 1], k=3, s=0)(self.rho)
                na = ne * frac_a
            except (AttributeError, TypeError):
                na = np.zeros(self.rho.shape)

        nn = namedtuple('nn', 's t tot')(
            ne * 1E-7,  # slow
            ne * 1E-7,  # thermal
            2 * ne * 1E-7  # total
        )
        self.n = namedtuple('n', 'D T i e C W Be Ne Ar Kr a n')(nD, nT, ni, ne, nC, nW, nBe, nNe, nAr, nKr, na, nn)

        self.z_0 = nC * 6.0 ** 2 / (nD + nT)  # TODO: Update this calculation
        self.z_eff = ((nD + nT) * 1.0 ** 2 + nC * 6.0 ** 2) / ne  # TODO: Update this calculation

        # populate temperature namedtuples, including the main T object
        Ti_kev_fit = UnivariateSpline(
            inp.Ti_data[:, 0],
            inp.Ti_data[:, 1],
            k=3,
            s=0.0
        )

        Ti_kev = Ti_kev_fit(self.rho)
        Te_kev = UnivariateSpline(inp.Te_data[:, 0], inp.Te_data[:, 1], k=3, s=0.0)(self.rho)
        Tns_kev = np.full(self.rho.shape, 0.002)
        Tnt_kev = Ti_kev
        try:
            TC_kev = UnivariateSpline(inp.TC_data[:, 0], inp.TC_data[:, 1], k=3, s=0.0)(self.rho)
        except:
            TC_kev = Ti_kev

        self.T = namedtuple('T', 'i e n C')(
            namedtuple('Ti', 'kev ev J')(
                Ti_kev,
                Ti_kev * 1E3,
                Ti_kev * 1E3 * e),
            namedtuple('eT', 'kev ev J')(
                Te_kev,
                Te_kev * 1E3,
                Te_kev * 1E3 * e),
            namedtuple('Tn', 's t')(
                namedtuple('Tn_s', 'kev ev J')(
                    Tns_kev,
                    Tns_kev * 1E3,
                    Tns_kev * 1E3 * e),
                namedtuple('Tn_t', 'kev ev J')(
                    Tnt_kev,
                    Tnt_kev * 1E3,
                    Tnt_kev * 1E3 * e)
            ),
            namedtuple('TC', 'kev ev J')(
                TC_kev,
                TC_kev * 1E3,
                TC_kev * 1E3 * e))

        # initialize pressures
        self.p = namedtuple('p', 'i e C')(
            self.n.i * self.T.i.J,
            self.n.e * self.T.e.J,
            self.n.C * self.T.C.J)

        # initialize spatial gradients and gradient scale lengths
        self.dni_dr = calc_grad(self.rho, self.n.i, self.psi_norm, self.R, self.Z, self.psi_data)
        # plt.contourf(self.R, self.Z, dni_dr, 500)
        # plt.axis('equal')
        # plt.colorbar()
        # plt.show()
        # sys.exit()
        self.dne_dr = calc_grad(self.rho, self.n.e, self.psi_norm, self.R, self.Z, self.psi_data)
        self.dnC_dr = calc_grad(self.rho, self.n.C, self.psi_norm, self.R, self.Z, self.psi_data)
        self.dTi_kev_dr = calc_grad(self.rho, self.T.i.kev, self.psi_norm, self.R, self.Z, self.psi_data)
        self.dTi_ev_dr = self.dTi_kev_dr * 1E3
        self.dTi_J_dr = self.dTi_kev_dr * 1E3 * e
        self.dTe_kev_dr = calc_grad(self.rho, self.T.i.kev, self.psi_norm, self.R, self.Z, self.psi_data)
        self.dTe_ev_dr = self.dTe_kev_dr * 1E3
        self.dTe_J_dr = self.dTe_kev_dr * 1E3 * e

        self.dpi_dr = calc_grad(self.rho, self.n.i * self.T.i.J, self.psi_norm, self.R, self.Z, self.psi_data)
        # Nick: Added Carbon Pressure gradient
        self.dpC_dr = calc_grad(self.rho, self.n.C * self.T.C.J, self.psi_norm, self.R, self.Z, self.psi_data)
        self.dpe_dr = calc_grad(self.rho, self.n.e * self.T.e.J, self.psi_norm, self.R, self.Z, self.psi_data)

        self.L = namedtuple('L', 'n T p')(
            namedtuple('Ln', 'i e')(
                -self.n.i / self.dni_dr,
                -self.n.e / self.dne_dr
            ),
            namedtuple('LT', 'i e')(
                -self.T.i.kev / self.dTi_kev_dr,  # note: independent of units,
                -self.T.e.kev / self.dTe_kev_dr  # note: independent of units
            ),
            namedtuple('Lp', 'i e')(
                -self.p.i / self.dpi_dr,  # note: independent of units,
                -self.p.e / self.dpe_dr  # note: independent of units
            )
        )

        try:
            try:
                E_r_fit = UnivariateSpline(inp.er_data[:, 0], inp.er_data[:, 1] * inp.Er_scale, k=3, s=0)
            except (AttributeError, TypeError):
                E_r_fit = UnivariateSpline(inp.er_data[:, 0], inp.er_data[:, 1], k=3, s=0)
            self.E_r = E_r_fit(self.rho)
            self.E_pot = np.zeros(self.rho.shape)
            try:
                for i, rhoval in enumerate(rho1d):
                    self.E_pot[i] = E_r_fit.integral(rhoval, self.sep_val)
            except:
                print "Error in E_pot integration"
        except:
            print 'Er data not supplied. Setting E_r and E_pot to zero.'
            self.E_r = np.zeros(self.rho.shape)
            self.E_pot = np.zeros(self.rho.shape)
        # initialize E_r and the corresponding electric potential

        # initialize rotation velocities from data
        try:
            vpolD = UnivariateSpline(inp.vpolD_data[:, 0], inp.vpolD_data[:, 1], k=3, s=0)(self.rho)
        except (AttributeError, TypeError):
            vpolD = np.zeros(self.rho.shape)
        try:
            vpolC = UnivariateSpline(inp.vpolC_data[:, 0], inp.vpolC_data[:, 1], k=3, s=0)(self.rho)
        except (AttributeError, TypeError):
            vpolC = np.zeros(self.rho.shape)
        try:
            vtorD = UnivariateSpline(inp.vtorD_data[:, 0], inp.vtorD_data[:, 1], k=3, s=0)(self.rho)
        except (AttributeError, TypeError):
            vtorD = np.zeros(self.rho.shape)
        try:
            vtorC = UnivariateSpline(inp.vtorC_data[:, 0], inp.vtorC_data[:, 1], k=3, s=0)(self.rho)
        except (AttributeError, TypeError):
            vtorC = np.zeros(self.rho.shape)

        self.v = namedtuple('v_comp', 'pol tor')(
            namedtuple('v_spec', 'D C')(vpolD, vpolC),
            namedtuple('v_spec', 'D C')(vtorD, vtorC),
        )

        self.v_1D = namedtuple('v_comp', 'pol tor')(
            namedtuple('v_spec', 'D C')(vpolD[:, 0], vpolC[:, 0]),
            namedtuple('v_spec', 'D C')(vtorD[:, 0], vtorC[:, 0]),
        )

        # initialize magnetic field-related quantities
        B_pol_raw = np.sqrt((np.gradient(self.psi_data.psi, self.psi_data.R[0], axis=1) / self.psi_data.R) ** 2 +
                            (np.gradient(self.psi_data.psi, self.psi_data.Z.T[0], axis=0) / self.psi_data.R) ** 2)

        self.B_p = griddata(np.column_stack((raw_psi_R.flatten(), raw_psi_Z.flatten())),
                            B_pol_raw.flatten(),
                            (self.R, self.Z),
                            method='linear')

        self.B_t = inp.BT0 * self.pts.axis.mag[0] / self.R
        self.B_tot = np.sqrt(self.B_p ** 2 + self.B_t ** 2)
        self.f_phi = self.B_t / self.B_tot

        try:
            # use input q profile if given
            q_interp = UnivariateSpline(inp.q_data[:, 0], inp.q_data[:, 1], k=3, s=1.0)
            self.q = q_interp(self.rho)
            self.q_1D = q_interp(self.rho.T[0])
        except:
            # otherwise calculate q-profile from psi data
            self.q_1D = inp.BT0 * self.pts.axis.mag[0] / (2 * pi) * calc_fs_perim_int(1.0 / (self.R ** 2 * self.B_p),
                                                                                      self.R, self.Z)
            self.q_1D[0] = self.q_1D[1]
            self.q = np.repeat(self.q_1D[np.newaxis, :], self.rho.shape[1], axis=0).T
        self.q0 = self.q_1D[0]
        self.q95 = UnivariateSpline(self.psi_norm[:, 0], self.q_1D, k=1, s=0)(0.95)

        # create Lz-related variables. These will remain zero unless set by the ImpRad module
        self.Lz_C = namedtuple('Lz_C', 's t ddT')(
            np.zeros(self.rho.shape),
            np.zeros(self.rho.shape),
            namedtuple('Lz_ddT', 's t')(
                np.zeros(self.rho.shape),
                np.zeros(self.rho.shape)
            )
        )

        self.Lz_Be = namedtuple('Lz_Be', 's t ddT')(
            np.zeros(self.rho.shape),
            np.zeros(self.rho.shape),
            namedtuple('Lz_ddT', 's t')(
                np.zeros(self.rho.shape),
                np.zeros(self.rho.shape)
            )
        )

        self.Lz_W = namedtuple('Lz_W', 's t ddT')(
            np.zeros(self.rho.shape),
            np.zeros(self.rho.shape),
            namedtuple('Lz_ddT', 's t')(
                np.zeros(self.rho.shape),
                np.zeros(self.rho.shape)
            )
        )

        self.Lz_Ne = namedtuple('Lz_Ne', 's t ddT')(
            np.zeros(self.rho.shape),
            np.zeros(self.rho.shape),
            namedtuple('Lz_ddT', 's t')(
                np.zeros(self.rho.shape),
                np.zeros(self.rho.shape)
            )
        )

        self.Lz_Ar = namedtuple('Lz_Ar', 's t ddT')(
            np.zeros(self.rho.shape),
            np.zeros(self.rho.shape),
            namedtuple('Lz_ddT', 's t')(
                np.zeros(self.rho.shape),
                np.zeros(self.rho.shape)
            )
        )

        self.Lz_Kr = namedtuple('Lz_Kr', 's t ddT')(
            np.zeros(self.rho.shape),
            np.zeros(self.rho.shape),
            namedtuple('Lz_ddT', 's t')(
                np.zeros(self.rho.shape),
                np.zeros(self.rho.shape)
            )
        )

        # calculate cross sections on the main computational grid
        svfus_dd, svfus_dd_ddT = calc_svfus(self.T, mode='dd')
        svfus_dt, svfus_dt_ddT = calc_svfus(self.T, mode='dt')
        svrec_st, svrec_st_ddT = calc_svrec_st(self.n, self.T)
        svcx_st, svcx_st_ddT = calc_svcx_st(self.T)
        svion_st, svion_st_ddT = calc_svion_st(self.T)
        svel_st, svel_st_ddT = calc_svel_st(self.T)

        self.sv = namedtuple('sv', 'fus rec cx ion el')(
            namedtuple('sv_fus', 'dd dt d_dT')(
                svfus_dd,
                svfus_dt,
                namedtuple('sv_fus_d_dT', 'dd dt')(
                    svfus_dd_ddT,
                    svfus_dt_ddT)),
            namedtuple('sv_rec', 'st d_dT')(
                svrec_st,
                namedtuple('sv_rec_d_dT', 'st')(
                    svrec_st_ddT)),
            namedtuple('sv_cx', 'st d_dT')(
                svcx_st,
                namedtuple('sv_cx_d_dT', 'st')(
                    svcx_st_ddT)),
            namedtuple('sv_ion', 'st d_dT')(
                svion_st,
                namedtuple('sv_ion_d_dT', 'st')(
                    svion_st_ddT)),
            namedtuple('sv_el', 'st d_dT')(
                svel_st,
                namedtuple('sv_el_d_dT', 'st')(
                    svel_st_ddT)))

        # Calculate some 1D Flux-surface averaged quantities
        self.izn_rate_fsa_s = calc_fsa(self.izn_rate.s, self.R, self.Z)
        self.izn_rate_fsa_t = calc_fsa(self.izn_rate.t, self.R, self.Z)
        self.izn_rate_fsa = calc_fsa(self.izn_rate.s + self.izn_rate.t, self.R, self.Z)
        self.cool_rate_fsa = calc_fsa(self.cool_rate, self.R, self.Z)
        self.dn_dr = calc_grad(self.rho, self.n.n.s + self.n.n.t, self.psi_norm, self.R, self.Z, self.psi_data)
        self.z_eff_fsa = calc_fsa(self.z_eff, self.R, self.Z)
        self.B_p_fsa = calc_fsa(self.B_p, self.R, self.Z)
        self.B_t_fsa = calc_fsa(self.B_t, self.R, self.Z)
        self.E_r_fsa = calc_fsa(self.E_r, self.R, self.Z)  # Piper changes: Also get FSA'd Er.

        self.n_fsa = namedtuple('n', 'i e n C')(
            calc_fsa(self.n.i, self.R, self.Z),
            calc_fsa(self.n.e, self.R, self.Z),
            namedtuple('nn', 's t tot')(
                calc_fsa(self.n.n.s, self.R, self.Z),  # slow
                calc_fsa(self.n.n.t, self.R, self.Z),  # thermal
                calc_fsa(self.n.n.tot, self.R, self.Z)  # total
            ),
            calc_fsa(self.n.C, self.R, self.Z))

        Ti_kev_fsa = calc_fsa(self.T.i.kev, self.R, self.Z)
        Te_kev_fsa = calc_fsa(self.T.e.kev, self.R, self.Z)
        Tns_kev_fsa = calc_fsa(self.T.n.s.kev, self.R, self.Z)
        Tnt_kev_fsa = calc_fsa(self.T.n.t.kev, self.R, self.Z)
        TC_kev_fsa = calc_fsa(self.T.C.kev, self.R, self.Z)

        self.T_fsa = namedtuple('T', 'i e n C')(
            namedtuple('Ti', 'kev ev J')(
                Ti_kev_fsa,
                Ti_kev_fsa * 1E3,
                Ti_kev_fsa * 1E3 * e),
            namedtuple('Te', 'kev ev J')(
                Te_kev_fsa,
                Te_kev_fsa * 1E3,
                Te_kev_fsa * 1E3 * e),
            namedtuple('Tn', 's t')(
                namedtuple('Tn_s', 'kev ev J')(
                    Tns_kev_fsa,
                    Tns_kev_fsa * 1E3,
                    Tns_kev_fsa * 1E3 * e),
                namedtuple('Tn_t', 'kev ev J')(
                    Tnt_kev_fsa,
                    Tnt_kev_fsa * 1E3,
                    Tnt_kev_fsa * 1E3 * e)
            ),
            namedtuple('TC', 'kev ev J')(
                TC_kev_fsa,
                TC_kev_fsa * 1E3,
                TC_kev_fsa * 1E3 * e))
        # Piper Changes: Calculate FSA of pressures and make namedtuple.
        pi_fsa = calc_fsa(self.n.i * self.T.i.J, self.R, self.Z)
        pC_fsa = calc_fsa(self.n.C * self.T.C.J, self.R, self.Z)
        pe_fsa = calc_fsa(self.n.e * self.T.e.J, self.R, self.Z)

        self.p_fsa = namedtuple('p_fsa', 'i C e')(
            -pi_fsa,
            -pC_fsa,
            -pe_fsa)

        # Piper Changes: Calculate FSA of pressure gradient.
        dpi_dr_fsa = calc_fsa(self.dpi_dr, self.R, self.Z)
        dpC_dr_fsa = calc_fsa(self.dpC_dr, self.R, self.Z)
        dpe_dr_fsa = calc_fsa(self.dpe_dr, self.R, self.Z)

        self.dp_dr_fsa = namedtuple('dp_dr_fsa', 'i C e')(
            -dpi_dr_fsa,
            -dpC_dr_fsa,
            -dpe_dr_fsa)

        # FSA of gradient scale length.
        self.Lp_fsa = namedtuple('Lp_fsa', 'i C e')(
            -dpi_dr_fsa / pi_fsa,
            -dpC_dr_fsa / pC_fsa,
            -dpe_dr_fsa / pe_fsa)

        self.L_fsa = namedtuple('L', 'T n p')(
            namedtuple('T', 'i e')(
                calc_fsa(self.L.T.i, self.R, self.Z),
                calc_fsa(self.L.T.e, self.R, self.Z)),
            namedtuple('n', 'i e')(
                calc_fsa(self.L.n.i, self.R, self.Z),
                calc_fsa(self.L.n.e, self.R, self.Z)),
            namedtuple('p', 'i e')(
                calc_fsa(self.L.p.i, self.R, self.Z),
                calc_fsa(self.L.p.e, self.R, self.Z))
        )
        # initialize chi_r using Bohm diffusion. This might get updated later
        chi_bohm = 5 / 32 * self.T.i.J / (e * self.B_tot)

        # calculate chi using the JET Bohm-GyroBohm model
        chi_i_jet, chi_e_jet = calc_chi_jet(self.T, self.L, self.a, self.q, self.B_t, m_d, self.rho)

        # chi hybrid
        # this chi uses the JET chi up to rho = 0.9 and then linearly interpolates to the value of chi_bohm
        # at the seperatrix

        self.chi = namedtuple('chi', 'bohm jet')(
            chi_bohm,
            namedtuple('jet', 'i e')(
                chi_i_jet,
                chi_e_jet
            )
        )
        # TODO: Verify this fus_rate calculation. I think it's wrong.
        self.fus_rate = 1 / 4 * self.n.D * self.n.D * self.sv.fus.dd + 1 / 4 * self.n.D * self.n.T * self.sv.fus.dt
        plt.clf()
        plt.close()

    def plot_te(self, edge=False):
        """
        Plots the GT3.Core electron temperature
        """
        fig = self._plot_base(self.T_fsa.e.kev,yLabel=r'$T_e [keV]$', title="GT3.Core Electron Temperature", edge=edge)
        return fig

    def _plot_base(self, val, xLabel=r'$\rho$', yLabel="Value", title="Title", color='red', edge=False):

        plot = plt.figure()
        fig = plot.add_subplot(111)
        fig.set_xlabel(xLabel, fontsize=20)
        fig.set_ylabel(yLabel, fontsize=20)
        fig.set_title(title)
        if edge:
            fig.set_xlim(0.85, self.sep_val)
        fig.scatter(self.rho[:, 0], val, color=color, s=MARKERSIZE)
        plt.show()
        return fig

    def _plot_with_wall(self):
        """
        Generates a Matplotlib plot with the wall pre-plotted.

        :return: An Axis object with the wall line plotted
        """
        fig_width = 6.0
        # Check to see if self.Z/R have been defind yet to generate a figure height, as they aren't immediately
        # calculated
        try:
            fig_height = (np.amax(self.Z) - np.amin(self.Z)) / (np.amax(self.R) - np.amin(self.R)) * fig_width
        except:
            fig_height = 9.0

        fig = plt.figure(figsize=(0.975*fig_width, fig_height))
        ax1 = fig.add_subplot(1, 1, 1)
        ax1.axis('equal')
        ax1.plot(np.asarray(self.wall_line.xy)[0], np.asarray(self.wall_line.xy)[1], color='black', lw=1.5)
        return ax1

    def _shapely_obj_plot_hanlder(self, obj, ax):
        if isinstance(obj, Point):
            ax.scatter(obj.x, obj.y, color='red', marker='o', s=MARKERSIZE)
            return ax
        if isinstance(obj, MultiPoint):
            for p in obj:
                ax.scatter(p.x, p.y, color='red', marker='o', s=MARKERSIZE)
            return ax
        if isinstance(obj, Path):
            ls = LineString(obj.vertices)
            ax.plot(*ls.xy, color='red', marker='o', s=MARKERSIZE)
            return ax
        if isinstance(obj, LineString):
            ax.plot(*obj.xy, color='red', marker='o', s=MARKERSIZE)
            return ax

    def _unknown_data_plot_helper(self, obj, ax):
        # Is it a shapely or similar object?
        if isinstance(obj, (Point, MultiPoint, Path, LineString)):
            return self._shapely_obj_plot_hanlder(obj, ax)

        # Does it have the vertices property to convert to a LineString?
        try:
            ls = LineString(obj.vertices)
            return self._shapely_obj_plot_hanlder(ls, ax)
        except:
            pass

        # Does it have an xy property?

        try:
            ax.plot(*obj.xy, s=MARKERSIZE)
            return ax
        except:
            pass

        # Can I turn it into a point?
        try:
            p = Point(obj)
            return self._shapely_obj_plot_hanlder(p, ax)
        except:
            pass

        # Can I turn it into a LineString directly?
        try:
            ls = LineString(obj)
            return self._shapely_obj_plot_hanlder(ls, ax)
        except:
            pass

        # No dice. Raising error.
        raise

    def plot_with_wall(self, obj=None):
        ax = self._plot_with_wall()

        try:
            return self._unknown_data_plot_helper(obj, ax)
        except:
            pass

        # Is it an iterable?
        try:
            obj.__iter__
            for p in obj:
                try:
                    ax = self._unknown_data_plot_helper(p, ax)
                except:
                    raise
            return ax
        except:
            pass

        print "Could not plot given data"

    def plot_exp_psi(self, res=50):
        ax = self._plot_with_wall()
        try:
            ax.contour(self.R, self.Z, self.psi, res)
            return ax
        except NameError:
            print "Psi not defined"
            pass

    def plot_L_t_i(self, edge=False):
        """
        Plots the GT3.Core ion temperature scale length
        """
        fig = self._plot_base(self.L.T.i[:,0], yLabel=r'$L_{T,i} [m]$', title="GT3.Core Ion Temperature Scale Length", edge=edge)
        return fig

    def plot_L_t_e(self, edge=False):
        """
        Plots the GT3.Core electron temperature scale length
        """
        fig = self._plot_base(self.L.T.e[:,0], yLabel=r'$L_{T,e} [m]$', title="GT3.Core Electron Temperature Scale Length", edge=edge)
        return fig

    def plot_L_n_i(self, edge=False):
        """
        Plots the GT3.Core ion density scale length
        """
        fig = self._plot_base(self.L.n.i[:,0], yLabel=r'$L_{n,i} [m]$', title="GT3.Core Ion Density Scale Length", edge=edge)
        return fig

    def plot_L_n_e(self, edge=False):
        """
        Plots the GT3.Core electron density scale length
        """
        fig = self._plot_base(self.L.n.e[:,0], yLabel=r'$L_{n,e} [m]$', title="GT3.Core Electron Density Scale Length", edge=edge)
        return fig

    def plot_ti(self, edge=False):
        """
        Plots the GT3.Core ion temperature
        """
        fig = self._plot_base(self.T_fsa.e.kev, yLabel=r'$T_i [keV]$', title="GT3.Core Ion Temperature", edge=edge)
        return fig

    def plot_t(self, edge=False):
        """
        Plots the 1D GT3.Core ion and electron temperatures
        """
        fig = self._plot_base(self.T_fsa.e.kev, yLabel=r'$T [keV]$', title="GT3.Core ion/electron temperatures", edge=edge)
        fig.scatter(self.rho[:, 0], self.T_fsa.i.kev, color="blue", s=MARKERSIZE)
        fig.legend([r'$T_e$', r'$T_i$'])
        return fig

    def plot_ni(self, edge=False):
        """
        Plots the 1D GT3.Core ion density
        """
        fig = self._plot_base(self.n_fsa.i, yLabel=r'$n_i [#/m^3]$', title="GT3.Core Ion density", edge=edge)
        return fig

    def plot_ne(self, edge=False):
        """
        Plots the 1D GT3.Core electron density
        """
        fig = self._plot_base(self.n_fsa.e, yLabel=r'$n_e [#/m^3]$', title="GT3.Core electron density", edge=edge)
        return fig

    def plot_n(self, edge=False):
        """
        Plots the 1D GT3.Core ion and electron densities
        """
        fig = self._plot_base(self.n_fsa.e, yLabel=r'$n [#/m^3]$', title="GT3.Core ion/electron densities", edge=edge)
        fig.scatter(self.rho[:, 0], self.n_fsa.i, color="blue", s=MARKERSIZE)
        fig.legend([r'$n_e$', r'$n_i$'])
        return fig

    def plot_ntrl_s(self, edge=False):
        """
        Plots the 1D GT3.Core slow neutral density
        """
        fig = self._plot_base(self.n_fsa.n.s, yLabel=r'$n_{o,s} [#/m^3]$', title="GT3.Core Slow Neutral Density", edge=edge)
        return fig

    def plot_ntrl_s(self, edge=False):
        """
        Plots the 1D GT3.Core thermal neutral density
        """
        fig = self._plot_base(self.n_fsa.n.t, yLabel=r'$n_{o,t} [#/m^3]$', title="GT3.Core Thermal Neutral Density", edge=edge)
        return fig

    def plot_ntrl_total(self, edge=False):
        """
        Plots the 1D GT3.Core total neutral density
        """
        fig = self._plot_base(self.n_fsa.n.tot, yLabel=r'$n_{o,tot} [#/m^3]$', title="GT3.Core Total Neutral Density", edge=edge)
        return fig

    def plot_er(self, edge=False):
        """
        Plots the 1D GT3.Core electric field
        """
        fig = self._plot_base(self.E_r_fsa, yLabel=r'$E_r [V/m]$', title="GT3.Core Radial Electric Field",
                              edge=edge)
        return fig

    def plot_vpol_C(self, edge=False):
        """
        Plots the 1D GT3.Core carbon poloidal velocity
        """

        fig = self._plot_base(self.v_1D.pol.C, yLabel=r'$V_{C,\theta} [km/s]$',
                              title="GT3.Core Carbon Poloidal Rotation Velocity",
                              edge=edge)
        return fig

    def plot_vpol_D(self, edge=False):
        """
        Plots the 1D GT3.Core deuterium poloidal velocity
        """

        fig = self._plot_base(self.v_1D.pol.D, yLabel=r'$V_{C,\theta} [km/s]$',
                              title="GT3.Core Deuterium Poloidal Rotation Velocity",
                              edge=edge)
        return fig

    def plot_vtor_C(self, edge=False):
        """
        Plots the 1D GT3.Core carbon toroidal velocity
        """
        fig = self._plot_base(self.v_1D.tor.C, yLabel=r'$V_{c,\phi} [km/s]$',
                              title="GT3.Core Carbon toroidal velocity",
                              edge=edge)
        return fig

    def plot_vtor_D(self, edge=False):
        """
        Plots the 1D GT3.Core deuterium toroidal velocity
        """

        fig = self._plot_base(self.v_1D.tor.D, yLabel=r'$V_{c,\phi} [km/s]$',
                              title="GT3.Core deuterium toroidal velocity",
                              edge=edge)
        return fig

    def plot_pressure_C(self, edge=False):
        """
        Plots the 1D GT3.Core carbon pressure
        """

        fig = self._plot_base(self.p_fsa.C, yLabel=r'$P_C [??]$',
                              title="GT3.Core carbon pressure",
                              edge=edge)
        return fig

    def plot_pressure_D(self, edge=False):
        """
        Plots the 1D GT3.Core deuterium pressure
        """

        fig = self._plot_base(self.p_fsa.i, yLabel=r'$P_i [??]$',
                              title="GT3.Core deuterium pressure",
                              edge=edge)
        return fig

    def plot_pressure_e(self, edge=False):
        """
        Plots the 1D GT3.Core electron pressure
        """

        fig = self._plot_base(self.p_fsa.e, yLabel=r'$P_e [??]$',
                              title="GT3.Core electron pressure",
                              edge=edge)
        return fig

    def plot_izn_rate_s(self, edge=False):
        """
        Plots the 1D GT3.Core slow neutral ionization rate
        """

        fig = self._plot_base(self.izn_rate_fsa_s, yLabel=r'$<\sigma v>_{slow} [??]$',
                              title="GT3.Core slow neutral ionization rate",
                              edge=edge)
        return fig

    def plot_izn_rate_t(self, edge=False):
        """
        Plots the 1D GT3.Core thermal neutral ionization rate
        """

        fig = self._plot_base(self.izn_rate_fsa_t, yLabel=r'$<\sigma v>_{thermal} [??]$',
                              title="GT3.Core thermal neutral ionization rate",
                              edge=edge)
        return fig

    def plot_izn_rate_total(self, edge=False):
        """
        Plots the 1D GT3.Core total neutral ionization rate
        """

        fig = self._plot_base(self.izn_rate_fsa, yLabel=r'$<\sigma v>_{tot} [??]$',
                              title="GT3.Core total neutral ionization rate",
                              edge=edge)
        return fig

    def plot_q(self, edge=False):
        """
        Plots the 1D GT3.Core safety factor
        """

        fig = self._plot_base(self.q_1D, yLabel=r'$q$',
                              title="GT3.Core safety factor",
                              edge=edge)
        return fig

    def update_ntrl_data(self, data):
        try:
            n_n_s = griddata(np.column_stack((data.R, data.Z)),
                             data.n_n_slow,
                             (self.R, self.Z),
                             method='linear')
        except:
            n_n_s = self.n.n.s

        try:
            n_n_t = griddata(np.column_stack((data.R, data.Z)),
                             data.n_n_thermal,
                             (self.R, self.Z),
                             method='linear')
        except:
            n_n_t = self.n.n.t

        try:
            izn_rate_s = griddata(np.column_stack((data.R, data.Z)),
                                  data.izn_rate_slow,
                                  (self.R, self.Z),
                                  method='linear')
        except:
            izn_rate_s = self.izn_rate.s

        try:
            izn_rate_t = griddata(np.column_stack((data.R, data.Z)),
                                  data.izn_rate_thermal,
                                  (self.R, self.Z),
                                  method='linear')
        except:
            izn_rate_t = self.izn_rate.t

        nn = namedtuple('nn', 's t tot')(
            n_n_s,  # slow
            n_n_t,  # thermal
            n_n_s + n_n_t  # total
        )

        self.n = namedtuple('n', 'i e n C')(self.n.i, self.n.e, nn, self.n.C)

        self.n_fsa = namedtuple('n', 'i e n C')(
            calc_fsa(self.n.i, self.R, self.Z),
            calc_fsa(self.n.e, self.R, self.Z),
            namedtuple('nn', 's t tot')(
                calc_fsa(self.n.n.s, self.R, self.Z),  # slow
                calc_fsa(self.n.n.t, self.R, self.Z),  # thermal
                calc_fsa(self.n.n.tot, self.R, self.Z)  # total
            ),
            calc_fsa(self.n.C, self.R, self.Z))

        self.izn_rate = namedtuple('izn_rate', 's t tot')(
            izn_rate_s,  # slow
            izn_rate_t,  # thermal
            izn_rate_s + izn_rate_t  # total
        )

        self.izn_rate_fsa_s = calc_fsa(self.izn_rate.s, self.R, self.Z)
        self.izn_rate_fsa_t = calc_fsa(self.izn_rate.t, self.R, self.Z)
        self.izn_rate_fsa = calc_fsa(self.izn_rate.s + self.izn_rate.t, self.R, self.Z)

        self.cool_rate_fsa = calc_fsa(self.n.e * self.n.C * np.nan_to_num(self.Lz_C.s) + \
                                      + self.n.e * self.n.C * np.nan_to_num(self.Lz_C.t),
                                      self.R, self.Z)
        self.dn_dr_fsa = calc_fsa(self.dn_dr, self.R, self.Z)

        # self.izn_rate_fsa = np.array(map(lambda x: self.izn_rate_fsa[-1] * x ** 10, self.r[:, 0] / self.a))
        # self.cool_rate_fsa = np.array(map(lambda x: self.cool_rate_fsa[-1] * x ** 10, self.r[:, 0] / self.a))
        # self.dn_dr_fsa = np.array(map(lambda x: self.dn_dr_fsa[-1] * x ** 10, self.r[:, 0] / self.a))

    def update_Lz_data(self, z, Lz, dLzdT):

        Lz_slow = Lz(np.log10(self.T.n.s.kev),
                     np.log10(self.n.n.s / self.n.e),
                     np.log10(self.T.e.kev))

        dLzdT_slow = dLzdT(np.log10(self.T.n.s.kev),
                           np.log10(self.n.n.s / self.n.e),
                           np.log10(self.T.e.kev))

        Lz_thermal = Lz(np.log10(self.T.n.t.kev),
                        np.log10(self.n.n.t / self.n.e),
                        np.log10(self.T.e.kev))

        dLzdT_thermal = dLzdT(np.log10(self.T.n.t.kev),
                              np.log10(self.n.n.t / self.n.e),
                              np.log10(self.T.e.kev))

        if z == 6:
            self.Lz_C = namedtuple('Lz_C', 's t ddT')(
                Lz_slow,
                Lz_thermal,
                namedtuple('Lz_ddT', 's t')(
                    dLzdT_slow,
                    dLzdT_thermal
                )
            )

        if z == 4:
            self.Lz_Be = namedtuple('Lz_Be', 's t ddT')(
                Lz_slow,
                Lz_thermal,
                namedtuple('Lz_ddT', 's t')(
                    dLzdT_slow,
                    dLzdT_thermal
                )
            )

        if z == 74:
            self.Lz_W = namedtuple('Lz_W', 's t ddT')(
                Lz_slow,
                Lz_thermal,
                namedtuple('Lz_ddT', 's t')(
                    dLzdT_slow,
                    dLzdT_thermal
                )
            )

        if z == 10:
            self.Lz_Ne = namedtuple('Lz_Ne', 's t ddT')(
                Lz_slow,
                Lz_thermal,
                namedtuple('Lz_ddT', 's t')(
                    dLzdT_slow,
                    dLzdT_thermal
                )
            )

        if z == 18:
            self.Lz_Ar = namedtuple('Lz_Ar', 's t ddT')(
                Lz_slow,
                Lz_thermal,
                namedtuple('Lz_ddT', 's t')(
                    dLzdT_slow,
                    dLzdT_thermal
                )
            )

        if z == 36:
            self.Lz_Kr = namedtuple('Lz_Kr', 's t ddT')(
                Lz_slow,
                Lz_thermal,
                namedtuple('Lz_ddT', 's t')(
                    dLzdT_slow,
                    dLzdT_thermal
                )
            )

        self.cool_rate_fsa = calc_fsa(self.n.e * self.n.C * np.nan_to_num(self.Lz_C.s) + \
                                      + self.n.e * self.n.C * np.nan_to_num(self.Lz_C.t),
                                      self.R, self.Z)  # TODO: Figure out how to not run this for all Z in a normal shot



#!/usr/bin/python

from GT3.TestBase import TestClass
import unittest
from GT3 import gt3
import neutpy

class NeutralsUnitTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.testPlasma = gt3(preparedInput=TestClass())

class NeutralsTestCase(NeutralsUnitTest):
    @classmethod
    def setUpClass(cls):
        super(NeutralsTestCase, cls).setUpClass()
    def test_get_plasma(self):
        self.assertIsInstance(self.testPlasma, gt3)
    def test_gt3_has_objs(self):
        self.assertTrue(hasattr(self.testPlasma, "core"))
        self.assertTrue(hasattr(self.testPlasma, "inp"))
    def test_neutpy_has_gt3_method(self):
        self.assertTrue(hasattr(neutpy.neutrals, "from_gt3"))
    def test_neutpy_run(self):
        self.npi = neutpy.neutrals()
        self.npi.from_gt3(self.testPlasma.core, self.testPlasma.inp)
    def test_neutpy_has_nn(self):
        self.assertTrue(hasattr(self.npi.nn, "s"))
        self.assertTrue(hasattr(self.npi.nn, "t"))
        self.assertTrue(hasattr(self.npi.nn, "tot"))
    def test_neutpy_has_izn(self):
        self.assertTrue(hasattr(self.npi.izn_rate, "s"))
        self.assertTrue(hasattr(self.npi.izn_rate, "t"))
        self.assertTrue(hasattr(self.npi.izn_rate, "tot"))

if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from shapely.geometry import  Point
from math import pi

def getangle(p1, p2):
    if isinstance(p1, Point) and isinstance(p2, Point):
        p1 = [p1.coords.xy[0][0], p1.coords.xy[1][0]]
        p2 = [p2.coords.xy[0][0], p2.coords.xy[1][0]]
    p1 = np.asarray(p1)
    p1 = np.reshape(p1, (-1, 2))
    p2 = np.asarray(p2)
    p2 = np.reshape(p2, (-1, 2))
    theta = np.arctan2(p1[:, 1]-p2[:, 1], p1[:, 0]-p2[:, 0])
    theta_mod = np.where(theta<0, theta+pi, theta)  # makes it so the angle is always measured counterclockwise from the horizontal
    return theta



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from contours.quad import QuadContourGenerator
from GT3.Core.Functions.IsClose import isclose
from GT3.Core.Functions.DrawCounterLine import draw_contour_line
from collections import namedtuple
from shapely.geometry import LineString
import numpy as np

def draw_core_line(R, Z, psinorm, psi_val, sep_pts):
    # create contour generator
    c = QuadContourGenerator.from_rectilinear(R[0], Z[:, 0], psinorm)

    # draw contours with psi_val
    contours = c.contour(psi_val)

    if len(contours) == 0 and isclose(psi_val, 0, abs_tol=1E-9):
        # This is probably the magnetic axis
        fs_line = None
        fs_inn_pt = None
        fs_out_pt = None
        fs_top_pt = None
        fs_bot_pt = None
        fs_axis = None
    elif len(contours) == 0 and not isclose(psi_val, 0, abs_tol=1E-9):
        # This either means that either:
        #    A) psi is a value not present in the psi data or
        #    B) psi is very close to the magnetic axis, but is too small for the contours
        #       package to give us a contour. This can happen if you you have an very fine
        #       radial mesh in the vicinity of the magnetic axis.
        # Passing back None for now, but should probably raise an exception in the future  # TODO
        fs_line = None
        fs_inn_pt = None
        fs_out_pt = None
        fs_top_pt = None
        fs_bot_pt = None
        fs_axis = None
    else:
        if len(contours) == 1:
            # then we're definitely dealing with a surface inside the seperatrix
            x, y = draw_contour_line(R, Z, psinorm, psi_val, 0)
        else:
            # we need to find which of the surfaces is inside the seperatrix
            for j, line in enumerate(contours):
                x, y = draw_contour_line(R, Z, psinorm, psi_val, j)

                if (np.amax(x) < np.amax(sep_pts[:, 0]) and
                    np.amin(x) > np.amin(sep_pts[:, 0]) and
                    np.amax(y) < np.amax(sep_pts[:, 1]) and
                    np.amin(y) > np.amin(sep_pts[:, 1])):
                    # then it's an internal flux surface
                    break

        pts = np.column_stack((x, y))
        fs_line = LineString(pts)
        fs_out_pt = pts[np.argmax(pts, axis=0)[0]]
        fs_inn_pt = pts[np.argmin(pts, axis=0)[0]]
        fs_top_pt = pts[np.argmax(pts, axis=0)[1]]
        fs_bot_pt = pts[np.argmin(pts, axis=0)[1]]
        fs_axis = [(fs_out_pt[0]+fs_inn_pt[0])/2, (fs_out_pt[1]+fs_inn_pt[1])/2]

    fs_pts = namedtuple('fs_pts', 'inn out top bot axis')(
        fs_inn_pt,
        fs_out_pt,
        fs_top_pt,
        fs_bot_pt,
        fs_axis
    )
    return fs_line, fs_pts


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from math import sqrt, pi
from GT3.RadialTransport.Functions.CalcReducedMass import calc_reduced_mass
from GT3.RadialTransport.Functions.CalcCoulLog import calc_coul_log
from scipy.constants import constants

e = constants.elementary_charge
eps_0 = constants.epsilon_0


def calc_t90(m1, m2, z1, z2, n2, T_J):
    """

    :param m1:
    :param m2:
    :param z1:
    :param z2:
    :param n2:
    :param T_J:
    :return:
    """
    m_r = calc_reduced_mass(m1, m2)
    coul_log = calc_coul_log(z1, z2, T_J, n2)
    t_90 = 2*pi*sqrt(m_r) * eps_0**2 * (3*T_J)**1.5 / \
           (n2 * ((z1*e) * (z2 * e))**2 * coul_log) # Piper Changes: Put the denominator in parenthises.
    return t_90


#!/usr/bin/python

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
deps = ['matplotlib',
        'multiprocess',
        'scipy',
        'numpy',
        'pandas',
        'pathos',
        'Shapely',
        'contours',
        'enum34',
        'PyYAML',
        'texttable']

setuptools.setup(
    name="GT3",
    version="0.0.2",
    author="Maxwell D. Hill, Jonathan J. Roveto, Nicholas Piper",
    install_requires=deps,
    author_email="max.hill@pm.me, veto1024@gmail.com, doom@gatech.edu",
    description="GT3 - The Georgia Tech Tokamak Transport codebase",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gt-frc/gt3/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Physics"
    ],
    python_requires='>=2.7',
)

if __name__ == '__main__':
    pass



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from scipy.interpolate import UnivariateSpline
from scipy.interpolate import griddata


def calc_grad(rho, quant, psi, R, Z, psi_data):
    rho_vals = rho[:,0]
    psi_vals = psi[:,0]
    vals = quant[:,0]


    # For some reason, psi_vals will sometimes not be monotonically increasing, especially near the magnetic axis.
    # This prevents UnivariateSpline from working. To prevent this, we're just going to delete any points that are
    # lower than the previous point, prior to doing the fit. As long as the psi data isn't too wonky, this should be
    # fine.
    psivals_mi = []  # psivals that are monotonically increasing
    vals_mi = []  # surf_area corresponding to monotonically increasing psi
    for i, psi_val in enumerate(psi_vals):
        if i == 0:
            psivals_mi.append(0)
            vals_mi.append(0)
        elif psi_val > psivals_mi[-1]:
            psivals_mi.append(psi_vals[i])
            vals_mi.append(vals[i])

    psivals_mi = np.asarray(psivals_mi)
    vals_mi = np.asarray(vals_mi)


    # calculate values as function of psi and get psi derivative function
    #psi_fit = UnivariateSpline(psivals_mi, vals_mi, k=3, s=0)
    psi_fit = UnivariateSpline(psi_vals, vals, k=3, s=0)
    d_dpsi_fit = psi_fit.derivative()

    # get value of dval_dpsi on the main computational grid
    dval_dpsi = d_dpsi_fit(psi)

    # calculate dpsi_norm_dr everywhere on the main computational grid
    dpsi_dr = griddata(np.column_stack((psi_data.R.flatten(),
                                        psi_data.Z.flatten())),
                       psi_data.dpsidr.flatten(),
                       (R,Z),
                       method='linear')

    dval_dr = dval_dpsi * dpsi_dr

    return dval_dr


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np


def calc_en_src_nbi(beam, iol_adjusted=False, E_orb_nbi=None):
    """
    """

    # Piper Changes: verify this calculation. Assumes the same things as the particle source eq.

    try:
        # nbi energy source in Joules/sec. Surprisingly, no need to actually calculate anything here.
        en_src_nbi = sum([beam.dPdr.v1D.W[i] for i in range(len(beam.E.J))])

        # nbi energy source in Joules/(m^3 * sec)

        en_src_nbi_tot = sum([beam.dPdV.v1D.W[i] for i in range(len(beam.E.J))])

    except:
        en_src_nbi = beam.dPdr.v1D.W
        en_src_nbi_tot = beam.dPdV.v1D.W

    if iol_adjusted:
        en_src_nbi = en_src_nbi * (1 - E_orb_nbi)
        en_src_nbi_lost = en_src_nbi_tot * (E_orb_nbi)
        en_src_nbi_kept = en_src_nbi_tot * (1 - E_orb_nbi)
    else:
        en_src_nbi_lost = np.zeros(en_src_nbi_tot.shape)
        en_src_nbi_kept = en_src_nbi_tot

    return en_src_nbi, en_src_nbi_tot, en_src_nbi_lost, en_src_nbi_kept



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
import re, os
import numpy as np
from GT3.BeamDeposition.Functions.CalcPwrFrac import calc_pwr_frac
from scipy.interpolate import interp1d, UnivariateSpline


class read_nbi_outfile:

    def __init__(self, inp, core, index = False, indBeam = False):
        if index:
            with open(os.getcwd() + '/outbeams_%s.dat' % str(index), 'r') as f:
                for count, line in enumerate(f):
                    if line.startswith(" Total Absorbed Power"):
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            P_abs_tot = float(result)
                        except:
                            P_abs_tot = np.NaN

                    if line.startswith(" Total Lost Power"):
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            P_lst_tot = float(result)
                        except:
                            P_lst_tot = np.NaN

                    if line.startswith(" Total NB Driven Current"):
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            I_nbi_tot = float(result)
                        except:
                            I_nbi_tot = np.NaN

                    if line.startswith(" Total NBCD Efficiency"):
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            I_nbi_eff = float(result)
                        except:
                            I_nbi_eff = np.NaN

                    if line.startswith(" Total Beam Beta"):
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            Beta_nbi_tot = float(result)
                        except:
                            Beta_nbi_tot = np.NaN

                    if line.startswith(" Taus"):
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            Taus_nbi = float(result)
                        except:
                            Taus_nbi = np.NaN

                    if line.startswith(" Volp"):
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            Volp_nbi = float(result)
                        except:
                            Volp_nbi = np.NaN

                    if line.startswith(" Absorbed Power"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            P_abs_1 = float(result)
                        except:
                            P_abs_1 = np.NaN

                    if line.startswith(" Lost Power"):
                        # this will need to be modified for multiple beams

                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            P_lst_1 = float(result)
                        except:
                            P_lst_1 = np.NaN

                    if line.startswith(" NB driven current"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            I_nbi_1 = float(result)
                        except:
                            I_nbi_1 = np.NaN

                    if line.startswith(" NBCD efficiency"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            I_nbi_eff_1 = float(result)
                        except:
                            I_nbi_eff_1 = np.NaN

                    if line.startswith(" NBCD gamma"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            I_nbi_gam_1 = float(result)
                        except:
                            I_nbi_gam_1 = np.NaN

                    if line.startswith("    energy group 1"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            st_en1_1 = float(result)
                        except:
                            st_en1_1 = np.NaN

                    if line.startswith("    energy group 2"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            st_en2_1 = float(result)
                        except:
                            st_en2_1 = np.NaN

                    if line.startswith("    energy group 3"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            st_en3_1 = float(result)
                        except:
                            st_en3_1 = np.NaN

                    if line.startswith(" Total Beam-Target Fusion Power"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            fus_pwr_bt = float(result)
                        except:
                            fus_pwr_bt = np.NaN

                    if line.startswith(" Total Power to Charged Particles"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            cp_pwr_tot = float(result)
                        except:
                            cp_pwr_tot = np.NaN

                    if line.startswith(" Total DT Neutron Rate"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            rate_dt_n = float(result)
                        except:
                            rate_dt_n = np.NaN

                    if line.startswith(" Total DD Neutron Rate"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            rate_dd_n = float(result)
                        except:
                            rate_dd_n = np.NaN

            with open(os.getcwd() + '/outbeams_%s.dat' % str(index), 'r') as f:
                data = f.read().replace('\n', ' ')
        else:
            with open(os.getcwd() + '/outbeams.dat', 'r') as f:
                for count, line in enumerate(f):
                    if line.startswith(" Total Absorbed Power"):
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            P_abs_tot = float(result)
                        except:
                            P_abs_tot = np.NaN

                    if line.startswith(" Total Lost Power"):
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            P_lst_tot = float(result)
                        except:
                            P_lst_tot = np.NaN

                    if line.startswith(" Total NB Driven Current"):
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            I_nbi_tot = float(result)
                        except:
                            I_nbi_tot = np.NaN

                    if line.startswith(" Total NBCD Efficiency"):
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            I_nbi_eff = float(result)
                        except:
                            I_nbi_eff = np.NaN

                    if line.startswith(" Total Beam Beta"):
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            Beta_nbi_tot = float(result)
                        except:
                            Beta_nbi_tot = np.NaN

                    if line.startswith(" Taus"):
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            Taus_nbi = float(result)
                        except:
                            Taus_nbi = np.NaN

                    if line.startswith(" Volp"):
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            Volp_nbi = float(result)
                        except:
                            Volp_nbi = np.NaN

                    if line.startswith(" Absorbed Power"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            P_abs_1 = float(result)
                        except:
                            P_abs_1 = np.NaN

                    if line.startswith(" Lost Power"):
                        # this will need to be modified for multiple beams

                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            P_lst_1 = float(result)
                        except:
                            P_lst_1 = np.NaN

                    if line.startswith(" NB driven current"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            I_nbi_1 = float(result)
                        except:
                            I_nbi_1 = np.NaN

                    if line.startswith(" NBCD efficiency"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            I_nbi_eff_1 = float(result)
                        except:
                            I_nbi_eff_1 = np.NaN

                    if line.startswith(" NBCD gamma"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            I_nbi_gam_1 = float(result)
                        except:
                            I_nbi_gam_1 = np.NaN

                    if line.startswith("    energy group 1"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            st_en1_1 = float(result)
                        except:
                            st_en1_1 = np.NaN

                    if line.startswith("    energy group 2"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            st_en2_1 = float(result)
                        except:
                            st_en2_1 = np.NaN

                    if line.startswith("    energy group 3"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            st_en3_1 = float(result)
                        except:
                            st_en3_1 = np.NaN

                    if line.startswith(" Total Beam-Target Fusion Power"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            fus_pwr_bt = float(result)
                        except:
                            fus_pwr_bt = np.NaN

                    if line.startswith(" Total Power to Charged Particles"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            cp_pwr_tot = float(result)
                        except:
                            cp_pwr_tot = np.NaN

                    if line.startswith(" Total DT Neutron Rate"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            rate_dt_n = float(result)
                        except:
                            rate_dt_n = np.NaN

                    if line.startswith(" Total DD Neutron Rate"):
                        # this will need to be modified for multiple beams
                        result = re.match(r'.*= *((?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)?)|NaN).*', line).group(1)
                        try:
                            rate_dd_n = float(result)
                        except:
                            rate_dd_n = np.NaN

            with open(os.getcwd() + '/outbeams.dat', 'r') as f:
                data = f.read().replace('\n', ' ')

        result = re.match(r'.*hofr_3 *((?:(?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)? +)|(?:NaN +))+) +Pitch.*', data).group(1)
        array = np.reshape(np.asarray(result.split(), dtype=float), (-1, 4))
        rho_nbeams = array[:, 0]
        hofr_1 = array[:, 1]
        hofr_2 = array[:, 2]
        hofr_3 = array[:, 3]

        result = re.match(r'.*zeta_3 *((?:(?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)? +)|(?:NaN +))+) +rho.*', data).group(1)
        array = np.reshape(np.asarray(result.split(), dtype=float), (-1, 4))
        ptch_angl1 = array[:, 1]
        ptch_angl2 = array[:, 2]
        ptch_angl3 = array[:, 3]

        result = re.match(r'.*dA *((?:(?:[-\+]?\d*(?:.?\d+)?(?:[Ee][-\+]?\d+)? +)|(?:NaN +))+) *', data).group(1)
        array = np.reshape(np.asarray(result.split(), dtype=float), (-1, 9))
        jnbtot = array[:, 1]
        pNBe   = array[:, 2]  # in MW/m^3
        pNBi   = array[:, 3]  # in MW/m^3
        nbfast = array[:, 4]
        pressb = array[:, 5]
        pfusb  = array[:, 6]
        dvol   = array[:, 7]
        dA     = array[:, 8]

        #
        # H(r) is a strangely normalized power DENSITY deposition profile
        #
        #   H(r) = (V_tot / P_beam) * dP/dV(r)
        #
        # The two most useful quantities for GT3 are:
        #
        #   dP/dV(r) = H(r) * (P_beam / V_tot)
        #
        # and
        #
        #   dP/dr(r) = dP/dV(r) * dV/dr(r) = H(r) * (P_beam / V_tot) * dV/dr(r)
        #



        if index:
            pwr_frac = calc_pwr_frac(indBeam['ebeam'])
        else:
            pwr_frac = calc_pwr_frac(inp.ebeam)
        self.beam_pwr_1 = P_abs_tot * pwr_frac[0]  # atomic deuterium
        self.beam_pwr_2 = P_abs_tot * pwr_frac[1]  # molecular deuterium D2
        self.beam_pwr_3 = P_abs_tot * pwr_frac[2]  # molecular deuterium D3

        if index:
            self.beam_en_1 = indBeam['ebeam']
            self.beam_en_2 = indBeam['ebeam'] / 2
            self.beam_en_3 = indBeam['ebeam'] / 3
        else:
            self.beam_en_1 = inp.ebeam
            self.beam_en_2 = inp.ebeam / 2
            self.beam_en_3 = inp.ebeam / 3



        dPdV_1_interp = UnivariateSpline(rho_nbeams,
                                          hofr_1 * self.beam_pwr_1 / Volp_nbi,
                                          k=3,
                                          s=0)

        dPdV_2_interp = UnivariateSpline(rho_nbeams,
                                          hofr_2 * self.beam_pwr_2 / Volp_nbi,
                                          k=3,
                                          s=0)

        dPdV_3_interp = UnivariateSpline(rho_nbeams,
                                          hofr_3 * self.beam_pwr_3 / Volp_nbi,
                                          k=3,
                                          s=0)

        dPdr_1_interp = UnivariateSpline(rho_nbeams,
                                         (hofr_1 * dvol * self.beam_pwr_1 )/ (.02 * Volp_nbi),
                                         k=3,
                                         s=0)

        dPdr_2_interp = UnivariateSpline(rho_nbeams,
                                         hofr_2  * dvol * self.beam_pwr_2 / (.02 * Volp_nbi),
                                         k=3,
                                         s=0)

        dPdr_3_interp = UnivariateSpline(rho_nbeams,
                                         hofr_3  * dvol * self.beam_pwr_3 / (.02 * Volp_nbi),
                                         k=3,
                                         s=0)

        zeta_1_interp = UnivariateSpline(rho_nbeams,
                                         ptch_angl1,
                                         k=3,
                                         s=0)

        zeta_2_interp = UnivariateSpline(rho_nbeams,
                                         ptch_angl2,
                                         k=3,
                                         s=0)

        zeta_3_interp = UnivariateSpline(rho_nbeams,
                                         ptch_angl3,
                                         k=3,
                                         s=0)

        self.dPdV_1 = dPdV_1_interp(core.rho)
        self.dPdV_2 = dPdV_2_interp(core.rho)
        self.dPdV_3 = dPdV_3_interp(core.rho)

        self.dPdr_1 = dPdr_1_interp(core.rho)
        self.dPdr_2 = dPdr_2_interp(core.rho)
        self.dPdr_3 = dPdr_3_interp(core.rho)

        self.dPdV_1_1D = dPdV_1_interp(core.rho[:, 0])
        self.dPdV_2_1D = dPdV_2_interp(core.rho[:, 0])
        self.dPdV_3_1D = dPdV_3_interp(core.rho[:, 0])


        # Introducing a normalization constant that seems to be getting lost in this calculation somewhere
        # TODO: Figure out what the hell is this

        self.dPdr_1_1D = dPdr_1_interp(core.rho[:, 0])
        self.dPdr_2_1D = dPdr_2_interp(core.rho[:, 0])
        self.dPdr_3_1D = dPdr_3_interp(core.rho[:, 0])

        # calculate deposition profile-weighted averaged zeta
        self.zeta_1 = np.sum(zeta_1_interp(np.linspace(0,1,100)) *
                             (1/self.beam_pwr_1) *
                             dPdr_1_interp(np.linspace(0,1,100)))
        self.zeta_2 = np.sum(zeta_1_interp(np.linspace(0,1,100)) *
                             (1/self.beam_pwr_1) *
                             dPdr_1_interp(np.linspace(0,1,100)))
        self.zeta_3 = np.sum(zeta_1_interp(np.linspace(0,1,100)) *
                             (1/self.beam_pwr_1) *
                             dPdr_1_interp(np.linspace(0,1,100)))

        self.debug={"P_abs_tot" : P_abs_tot,
                    "P_lst_tot"  : P_lst_tot,
                    "I_nbi_tot" : I_nbi_tot,
                    "I_nbi_eff" : I_nbi_eff,
                    "Volp_nbi"  : Volp_nbi,
                    "P_abs_1" : P_abs_1,
                    "hofr_1" : hofr_1,
                    "hofr_2" : hofr_2,
                    "hofr_3" : hofr_3,
                    "jnbtot" : jnbtot,
                    "pNBe" : pNBe,
                    "pNBi" : pNBi,
                    "nbfast" : nbfast,
                    "pressb" : pressb,
                    "pfusb" : pfusb,
                    "dvol" : dvol,
                    "dA" : dA,
                    "rho_nbeams" : rho_nbeams}


f#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 16 14:58:56 2018

@author: max
"""
import matplotlib.pyplot as plt
from matplotlib import rc
#rc('font', **{'family':'sans-serif', 'sans-serif':['Helvetica']})
## for Palatino and other serif fonts use:
rc('font', **{'family':'serif', 'serif':['Palatino']})
rc('text', usetex=True)

class gt3plots():
    sys.dont_write_bytecode = True 
    #if ITER model
    #if myshot.param.d3d_iter==1:
    #    fig = plt.figure(figsize=(8, 8))
    #    ax1 = fig.add_subplot(111)
    #    ax1.set_title(r'ITER First Wall', fontsize=20, )
    #    ax1.set_ylabel(r'Z ($m$)', fontsize=20)
    #    ax1.set_xlabel(r'R ($m$)', fontsize=20)
    #    ax1.axis('equal')
    #    ax1.plot(myshot.exp.lim_ITER[:, 0], myshot.exp.lim_ITER[:, 1])
    
    #print ('test = ', myshot.ntrl.sol_flx_line[0])
    
    mill_plot = plt.figure(figsize=(12, 8))
    ax1 = mill_plot.add_subplot(111)
    ax1.axis('equal')
    ax1.set_xlim(3.5, 7)
    ax1.set_ylim(-4.5, -2.5)
    ax1.contour(myshot.brnd.R, myshot.brnd.Z, myshot.brnd.r, 30)
    ax1.plot(myshot.brnd.R[-1, :], myshot.brnd.Z[-1, :])
    sep = np.vstack((myshot.exp.bdry, myshot.exp.bdry[0]))
    #ax1.plot(myshot.ntrl.points_unique[:, 0], myshot.ntrl.points_unique[:, 1], 'o')
    #ax1.plot(myshot.ntrl.sol_pts[:, 0], myshot.ntrl.sol_pts[:, 1], color='black')
    for i in range(0, 5):
        #ax1.plot(myshot.ntrl.sol_flx_line[:, 2*i], myshot.ntrl.sol_flx_line[:, 2*i+1], color='black')
        ax1.plot(myshot.ntrl.sol_pts[i*100:(i+1)*100, 0], myshot.ntrl.sol_pts[i*100:(i+1)*100, 1], color='black')
    ax1.plot(myshot.exp.lim_vertex[:, 0], myshot.exp.lim_vertex[:, 1])
    #ax1.plot(myshot.ntrl.sol_lim_pts[:, 0], myshot.ntrl.sol_lim_pts[:, 1], 'o', color='red')
    #ax1.plot(sep[:, 0], sep[:, 1], color='green', lw=3)
    #CS = ax1.contourf(myshot.brnd.R, myshot.brnd.Z, myshot.brnd.B_p, 500) #plot something calculated by miller
    #plt.colorbar(CS)
    
    
    ITER_wall_x = np.array([4.04688, 4.04688, 4.30729, 4.98438, 5.73958, 7.38021, 7.88802, 
                            8.23958, 8.39583, 8.30469, 7.90104, 7.28906, 6.27344, 6.33854, 
                            6.27344, 6.15625, 6.05208, 5.96094, 5.86979, 5.77865, 5.71354, 
                            5.66146, 5.60938, 5.57031, 5.55729, 5.55729, 5.27083, 5.25781, 
                            5.25781, 5.20573, 5.15365, 5.07552, 4.95833, 4.85417, 4.6849, 
                            4.64583, 4.5026, 4.1901, 4.35938, 4.48958, 4.51562, 4.48958, 
                            4.42448, 4.33333, 4.21615, 4.125, 4.04688])
    ITER_wall_y = np.array([-2.56193, 3.60495, 4.28292, 4.66102, 4.53064, 3.16167, 2.43155, 
                            1.66232, 0.59322, -0.436767, -1.34941, -2.2751, -3.05737, 
                            -3.16167, -3.22686, -3.22686, -3.2399, -3.27901, -3.33116, 
                            -3.39635, -3.46154, -3.53977, -3.63103, -3.83963, -3.95698, 
                            -4.53064, -4.26988, -4.16558, -4.00913, -3.91786, -3.85267, 
                            -3.78748, -3.73533, -3.72229, -3.73533, -3.76141, -3.91786, 
                            -3.90482, -3.52673, -3.21382, -3.08344, -2.9661, -2.78357, 
                            -2.67927, -2.61408, -2.57497, -2.56193])
    
    R0 = myshot.param.R0_a
    a = myshot.param.a
    mill_plot2 = plt.figure(figsize=(8, 12))
    ax1 = mill_plot2.add_subplot(111)
    #ax1.set_title(r'Example Miller Flux Surfaces', fontsize=30)
    #ax1.set_ylabel(r'Z ($m$)', fontsize=30)
    #ax1.set_xlabel(r'R ($m$)', fontsize=30)
    #ax1.tick_params(labelsize=30)
    ax1.axis('equal')
    #print myshot.brnd.R.T[::-2]
    for i, vals in enumerate(myshot.brnd.R.T):
        if i%4 == 0:
            ax1.plot(myshot.brnd.R.T[i], myshot.brnd.Z.T[i], 'gray', alpha=0.5, lw=1)
    #ax1.plot(myshot.brnd.R[-1], myshot.brnd.Z[-1], 'black')
    CS1 = ax1.contour(myshot.brnd.R, myshot.brnd.Z, myshot.brnd.rho, levels = [0.333, 0.666, 0.999], colors='black', linewidths=6)
    
    fmt = {}
    strs = [r'$\rho = 0.33$', r'$\rho = 0.66$', r'$\rho = 1$']
    for l, s in zip(CS1.levels, strs):
        fmt[l] = s
    
    manual_locations = [(R0+a/3, 0.1), (R0+2*a/3, -0.2), (R0+a, -0.5)]
    # Label every other level using strings
    #plt.clabel(CS1, CS1.levels[::], inline=True, fmt=fmt, fontsize=10)
    
    labels = ax1.clabel(CS1, CS1.levels[::], inline=1, fmt=fmt, fontsize=20, manual = manual_locations)
    for l in labels:
        l.set_rotation(0)
    print 'myshot.param.d3d_iter = ', myshot.param.d3d_iter
    if myshot.param.d3d_iter==1:
        ax1.plot(ITER_wall_x, ITER_wall_y, lw=6)
    else:
        ax1.plot(myshot.exp.lim_vertex[:, 0], myshot.exp.lim_vertex[:, 1], lw=6)
    #ax1.plot(myshot.exp.lim_ITER[:, 0], myshot.exp.lim[:, 1], 'red')
    ax1.set_axis_off()
    mill_plot2.tight_layout()
    mill_plot2.savefig("/home/max/Nextcloud/Max/max_phd/Codes/X-MILLER/example_miller.png", dpi=300, transparent=True) 
    
    
    fig = plt.figure(figsize=(12, 8))
    fig.suptitle(r'Using original stuff')
    ax1 = fig.add_subplot(231)
    ax1.set_title(r'ni', fontsize=20)
    ax1.set_ylabel(r'ni', fontsize=20)
    ax1.set_xlabel(r'rho', fontsize=20)
    ax1.plot(myshot.brnd.rho[:, 0], myshot.brnd.ni[:, 0])
    
    ax2 = fig.add_subplot(232)
    ax2.set_title(r'ne', fontsize=20)
    ax2.set_ylabel(r'ne', fontsize=20)
    ax2.set_xlabel(r'rho', fontsize=20)
    ax2.plot(myshot.brnd.rho[:, 0], myshot.brnd.ne[:, 0])
    
    ax3 = fig.add_subplot(233)
    ax3.set_title(r'Ti', fontsize=20)
    ax3.set_ylabel(r'Ti', fontsize=20)
    ax3.set_xlabel(r'rho', fontsize=20)
    ax3.plot(myshot.brnd.rho[:, 0], myshot.brnd.Ti_kev[:, 0])
    
    ax4 = fig.add_subplot(234)
    ax4.set_title(r'Te', fontsize=20, )
    ax4.set_ylabel(r'Te', fontsize=20)
    ax4.set_xlabel(r'rho', fontsize=20)
    ax4.plot(myshot.brnd.rho[:, 0], myshot.brnd.Te_kev[:, 0])
    plt.tight_layout()
    
    ax4 = fig.add_subplot(235)
    ax4.set_title(r'E_pot', fontsize=20, )
    ax4.set_ylabel(r'E_pot', fontsize=20)
    ax4.set_xlabel(r'rho', fontsize=20)
    ax4.plot(myshot.brnd.rho[:, 0], myshot.brnd.E_pot[:, 0])
    plt.tight_layout()
    
    ax4 = fig.add_subplot(236)
    ax4.set_title(r'B_phi', fontsize=20, )
    ax4.set_ylabel(r'B_phi', fontsize=20)
    ax4.set_xlabel(r'R', fontsize=20)
    ax4.plot(myshot.brnd.R[:, myshot.brnd.R.shape[1]/2], myshot.brnd.B_phi[:, myshot.brnd.R.shape[1]/2], 'blue')
    ax4.plot(myshot.brnd.R[:, 0], myshot.brnd.B_phi[:, 0], 'blue')
    plt.tight_layout()
    
    #ax1.plot(myshot.brnd.R[-1, :], myshot.brnd.Z[-1, :])
    #sep = np.vstack((myshot.exp.bdry, myshot.exp.bdry[0]))
    #ax1.plot(sep[:, 0], sep[:, 1], color='green', lw=3)
    #CS = ax1.pcolor(myshot.brnd.R, myshot.brnd.Z, myshot.brnd.B_p, vmin=0, vmax=0.6) #plot something calculated by miller
    #CS = ax1.pcolor(myshot.brnd.R, myshot.brnd.Z, myshot.brnd.Psi_norm) #plot something calculated by miller
    #CS = ax1.pcolor(myshot.brnd.R, myshot.brnd.Z, B_p_in, )
    #plt.colorbar(CS)
    """
    """
    B_p_in = griddata(np.column_stack((myshot.exp.R.flatten(), myshot.exp.Z.flatten())), 
                            myshot.exp.B_p.flatten(), 
                            (myshot.brnd.R, myshot.brnd.Z), 
                            method='cubic')
    
    
    mill_plot3 = plt.figure(figsize=(8, 8))
    ax1 = mill_plot3.add_subplot(111)
    ax1.axis('equal')
    #ax1.contour(myshot.brnd.R, myshot.brnd.Z, myshot.brnd.r, 70)
    #ax1.plot(myshot.brnd.R[-1, :], myshot.brnd.Z[-1, :])
    #sep = np.vstack((myshot.exp.bdry, myshot.exp.bdry[0]))
    #ax1.plot(sep[:, 0], sep[:, 1], color='green', lw=3)
    CS = ax1.contourf(myshot.brnd.R, myshot.brnd.Z, myshot.brnd.B_p, 500)
    #CS = ax1.pcolor(myshot.brnd.R, myshot.brnd.Z, B_p_in, vmin=0, vmax=0.6)
    #CS = ax1.contourf(myshot.brnd.R, myshot.brnd.Z, B_p_in, 500, vmin=0, vmax=0.6) #plot something calculated by miller
    plt.colorbar(CS)
    
    
    
    """
    fig = plt.figure(figsize=(6, 6))
    #if myshot.param.d3d_iter==1:
        #fig.suptitle('Calculated IOL in ITER', fontsize=24, y=0.8)
    #else:
        #fig.suptitle('Calculated IOL in DIII-D', fontsize=24, y=0.8)
    ax1 = fig.add_subplot(111)
    #ax1.set_title(r'Calculated $E_{orb}$ in DIII-D', fontsize=20)
    ax1.set_title(r'Increase $E_{orb}$', fontsize=20)
    #ax1.set_ylabel(r'Cumulative energy loss fraction', fontsize=20)
    ax1.set_ylabel(r'Increase in cumulative energy loss', fontsize=20)
    ax1.set_xlabel(r'$\rho$', fontsize=20)
    #ax1.text(0.05, 0.85, 
    #    "Baseline Rad. = 53.492 MW\n$T_i$ increase Rad. = %s MW"%(round(rad_tot_adpak/1E6, 3)), 
    #    fontsize=15, 
    #    transform = ax1.transAxes, 
    #    bbox={'facecolor':'white', 'alpha':1, 'pad':10})
    
    ax1.grid(b=True, which='both', axis='both')
    ax1.set_xlim(0.75, 1)
    #ax1.set_ylim(0, 1)
    #ax1.plot(np.linspace(0, 1, myshot.param.rpts), myshot.tiol.E_orb, label='Eorb', color='black', lw=3)
    #ax1.plot(np.linspace(0, 1, myshot.param.rpts), E_orb_ITER, label='Baseline ITER', color='black', lw=3)
    #ax1.plot(np.linspace(0, 1, myshot.param.rpts), E_orb_ITER_T2, '--', label='20\% Temp increase', color='red', lw=3)
    
    ax1.plot(np.linspace(0, 1, myshot.param.rpts), (E_orb_ITER_T2 - E_orb_ITER), label='Increase', color='black', lw=3)
    
    ax1.legend(bbox_to_anchor=(0.015, 0.6), 
        loc='lower left', 
        ncol=1, 
        prop={'size':20}, 
        #transform = ax1.transAxes, 
        shadow=True)  
    """
    
    #fig2 = plt.figure(figsize=(6, 6))
    #if myshot.param.d3d_iter==1:
        #fig.suptitle('Calculated IOL in ITER', fontsize=24, y=0.8)
    #else:
        #fig.suptitle('Calculated IOL in DIII-D', fontsize=24, y=0.8)
    #ax1 = fig2.add_subplot(111)
    #ax1.contourf(myshot.brnd.R[:450, :], myshot.brnd.Z[:450, :], myshot.brnd.rho[:450, :], 1, colors='blue')
    #ax1.contourf(myshot.brnd.R[450:, :], myshot.brnd.Z[450:, :], myshot.brnd.rho[450:, :], 1, colors='green')
    #ax1.plot(myshot.brnd.R[-1, :], myshot.brnd.Z[-1, :], 'black', lw=3)
    #ax1.axis('equal')
    #plt.axis('off')
    
    
    """
    ax2 = fig.add_subplot(132)
    ax2.set_title(r'$M_{orb}$', fontsize=20)
    ax2.set_ylabel(r'cumulative momentum loss fraction', fontsize=20)
    ax2.set_xlabel(r'minor radius', fontsize=20)
    ax2.grid(b=True, which='both', axis='both')
    ax2.set_xlim(0.75, 1)
    ax2.plot(np.linspace(0, 1, myshot.param.rpts), myshot.tiol.M_orb, label='Morb', color='black', lw=3)
    
    ax3 = fig.add_subplot(133)
    ax3.set_title(r'$E_{orb}$', fontsize=20)
    ax3.set_ylabel(r'cumulative energy loss fraction', fontsize=20)
    ax3.set_xlabel(r'minor radius', fontsize=20)
    ax3.grid(b=True, which='both', axis='both')
    ax3.set_xlim(0.75, 1)
    ax3.plot(np.linspace(0, 1, myshot.param.rpts), myshot.tiol.E_orb, label='Eorb', color='black', lw=3)
    plt.tight_layout()
    fig.subplots_adjust(top=0.84)
    """
    """
    myshot = plasma('toplasma1')
    myshot.solve()
    np.savetxt('input1_fiol_t.dat', myshot.tiol.F_orb, delimiter=', ')
    np.savetxt('input1_miol_t.dat', myshot.tiol.M_orb, delimiter=', ')
    np.savetxt('input1_eiol_t.dat', myshot.tiol.E_orb, delimiter=', ')
    
    myshot = plasma('toplasma2')
    myshot.solve()
    np.savetxt('input2_fiol_t.dat', myshot.tiol.F_orb, delimiter=', ')
    np.savetxt('input2_miol_t.dat', myshot.tiol.M_orb, delimiter=', ')
    np.savetxt('input2_eiol_t.dat', myshot.tiol.E_orb, delimiter=', ')
    
    myshot = plasma('toplasma3')
    myshot.solve()
    np.savetxt('input3_fiol_t.dat', myshot.tiol.F_orb, delimiter=', ')
    np.savetxt('input3_miol_t.dat', myshot.tiol.M_orb, delimiter=', ')
    np.savetxt('input3_eiol_t.dat', myshot.tiol.E_orb, delimiter=', ')
    """


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
from math import  pi

def calc_fs_int(x, R, Z):
    R1 = R[:, :-1]
    R2 = np.roll(R[:, :-1], -1, axis=1)
    Z1 = Z[:, :-1]
    Z2 = np.roll(Z[:, :-1], -1, axis=1)
    x1 = x[:, :-1]
    x2 = np.roll(x[:, :-1], -1, axis=1)

    dl = np.sqrt((R2 - R1) ** 2 + (Z2 - Z1) ** 2)

    R_av = (R1 + R2)/2

    dA = dl * (2 * pi * R_av)

    x_av = (x1 + x2)/2

    fsa = np.sum(x_av * dA, axis=1)
    #fsa[0] = x[0,0]
    return fsa


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np


def calc_rho1d(edge_rho=None, rhopts_core=None, rhopts_edge=None, rhopts=None):
    # define rho points
    try:
        rho1d = np.concatenate((np.linspace(0, edge_rho, rhopts_core, endpoint=False),
                                np.linspace(edge_rho, 1, rhopts_edge, endpoint=True)))
    except:
        try:
            rho1d = np.linspace(0, 1, rhopts)
        except:
            print 'rho parameters not defined. Using 100 evenly spaced rho values'
            rho1d = np.linspace(0, 1, 100)
    return rho1d



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import numpy as np
import sys
from math import cos, sin
from shapely.geometry import Point, LineString
from GT3.Core.Functions.DrawCoreLine import draw_core_line


def calc_RZ(core, rho, theta, theta_xpt, pts, psi_data, psi_norm, lines):
    # get parameters that depend on both rho and theta

    sep_pts = np.asarray(lines.sep_closed.coords)

    R = np.zeros(rho.shape)
    Z = np.zeros(rho.shape)

    line_length = 5.0

    if pts.xpt[1] is None or pts.xpt[0] is None:
        # Grab the actual x-point
        if pts.xpt[0] is not None:
            xpt_temp = pts.xpt[0]
        else:
            xpt_temp = pts.xpt[1]
    else:
        # THere are 2 xpoints. Use the bottom one.
        xpt_temp = pts.xpt[0]

    for i, psi_norm_val in enumerate(psi_norm[:, 0]):
        if i == 0:
            R[i] = pts.axis.mag[0]
            Z[i] = pts.axis.mag[1]
        else:
            # attempt to draw flux surface line through that point
            # (may not work for flux surfaces very close to the magnetic axis)
            fs_line, fs_pts = draw_core_line(psi_data.R, psi_data.Z, psi_data.psi_norm, psi_norm_val, sep_pts)
            if fs_line is not None and fs_pts.axis is None:
                # then draw_core_line didn't find any contours, which probably means it was trying
                # to draw a surface closer to psi_norm=0 than the contours package would cooperate with.
                # When this happens, the best thing to do for now is decrease the radial resolution in
                # the vicinity of the magnetic axis. Ideas for the fixing this in the future:
                #   1) resample the raw psi data (maybe an Rbf interpolation) onto a finer mesh. May or may not work.
                #   2) some kind of interpolation based on the flux surfaces it can draw.  # TODO
                raise RuntimeError("""GT3 had trouble drawing a contour line when getting the R and Z points. This ' \
                      'is most likely due to an overly fine radial mesh in the vicnity of the magnetic ' \
                      'axis. Try reducing your number of radial meshes in the core and try again. This ' \
                      'will hopefully be fixed in a future update.""")

            for j, thetaval in enumerate(theta[0]):
                if psi_norm_val < 1.0:
                    thetaline = LineString([Point(fs_pts.axis),
                                            Point([line_length * cos(thetaval) + fs_pts.axis[0],
                                                   line_length * sin(thetaval) + fs_pts.axis[1]])])
                    int_pt = fs_line.intersection(thetaline)
                else:
                    if thetaval == theta_xpt:
                        int_pt = Point(xpt_temp)

                    else:
                        thetaline = LineString([Point(pts.axis.geo),
                                                Point([line_length * cos(thetaval) + pts.axis.geo[0],
                                                       line_length * sin(thetaval) + pts.axis.geo[1]])])
                        int_pt = LineString(sep_pts).intersection(thetaline)
                if isinstance(int_pt, Point):
                    R[i, j] = int_pt.x
                    Z[i, j] = int_pt.y
                else:
                    raise
    return R, Z


#!/usr/bin/python
from scipy import constants

"""Physics-related constants"""
epsilon_0 = constants.epsilon_0
elementary_charge = constants.elementary_charge
electron_mass = constants.electron_mass
deuteron_mass = constants.physical_constants['deuteron mass'][0]
N_A = constants.N_A
mu_0 = constants.mu_0
MARKERSIZE=6


#!/usr/bin/python

from TestBase.GT3TestBase import GT3TestBase
from GT3.gt3 import gt3

if __name__=="__main__":

    try:
        test = GT3TestBase()
        shot = gt3(preparedInput=test)
        shot.run_NBI()
        print "Test successful"
    except Exception as e:
        print "Test failed"
        raise e






# -*- coding: utf-8 -*-
"""
Created on Sat Aug  5 19:48:34 2017

@author: max
"""
from __future__ import division
import numpy as np
import matplotlib.pyplot as plt
from math import pi, sqrt
from collections import namedtuple
from Functions.CalcIOLMaxwellian import calc_iol_maxwellian
from Functions.CalcIOLMonoEn import calc_iol_mono_en
from Functions.CalcIOLBeams import calc_iol_beams
import sys

m_d = 3.343583719e-27
m_t = 5.006e-27
m_c = 1.9926467e-26
m_a = 6.643e-27


class IOL:
    """
    """

    def __init__(self, inp, core):
        sys.dont_write_bytecode = True
        np.warnings.filterwarnings('ignore')

        numcos = inp.numcos  # TODO: Clean this up
        angles = np.linspace(-1, 1, inp.numcos + 1)
        self.coslist = ((angles + np.roll(angles, -1)) / 2)[:-1]
        self.rho = core.rho
        self.calc_iol_beams = calc_iol_beams
        polpts = len(core.rho[-1])
        radpts = len(core.rho.T[-1])

        # THE FOLLOWING ARRAYS ARE 4-DIMENSIONAL ARRAYS
        # [ LAUNCH THETA POSITION , LAUNCH ANGLE COSINE, LAUNCH r  , EXIT THETA POSITION  ]

        # NOTE TO FUTURE DEVELOPERS: IF YOU TRY TO LOOP OVER THE PLASMA POINTS, LAUNCH ANGLES, AND
        # EXIT LOCATIONS IN PYTHON THE WAY YOU WOULD IN C OR FORTRAN, IT'S GOING TO TAKE FOREVER.
        # ALTHOUGH THESE ARRAYS TAKE MORE MEMORY THAT I'D LIKE, IT'S CURRENTLY NECESSARY TO DO IT THIS WAY.
        # MAYBE SOMETHING TO IMPROVE ON IN THE FUTURE.

        # CREATE ARRAYS FOR LAUNCH POINTS IN THE PLASMA
        r0 = np.broadcast_arrays(np.ones(polpts)[:, None, None, None],
                                 np.ones(inp.numcos)[:, None, None],
                                 core.rho)[-1]

        B0 = np.broadcast_arrays(np.ones(polpts)[:, None, None, None],
                                 np.ones(inp.numcos)[:, None, None],
                                 core.B_tot)[-1]

        f0 = np.broadcast_arrays(np.ones(polpts)[:, None, None, None],
                                 np.ones(inp.numcos)[:, None, None],
                                 core.f_phi)[-1]

        psi0 = np.broadcast_arrays(np.ones(polpts)[:, None, None, None],
                                   np.ones(inp.numcos)[:, None, None],
                                   core.psi)[-1]

        phi0 = np.broadcast_arrays(np.ones(polpts)[:, None, None, None],
                                   np.ones(inp.numcos)[:, None, None],
                                   core.E_pot)[-1]  # * 1E3  # now in volts

        zeta0 = np.broadcast_arrays(np.ones(polpts)[:, None, None, None],
                                    self.coslist[:, None, None],
                                    np.ones(core.R.shape))[1]

        # CREATE ARRAYS FOR DESTINATION POINTS ALONG THE SEPERATRIX
        R1 = np.broadcast_arrays(np.ones(polpts)[:, None, None, None],
                                 np.ones(inp.numcos)[:, None, None],
                                 np.ones(radpts)[:, None],
                                 np.ones(polpts)[:],
                                 core.R[-1][:, None, None, None])[-1]

        f1 = np.broadcast_arrays(np.ones(polpts)[:, None, None, None],
                                 np.ones(inp.numcos)[:, None, None],
                                 np.ones(radpts)[:, None], np.ones(polpts)[:],
                                 core.f_phi[-1][:, None, None, None])[-1]

        B1 = np.broadcast_arrays(np.ones(polpts)[:, None, None, None],
                                 np.ones(inp.numcos)[:, None, None],
                                 np.ones(radpts)[:, None],
                                 np.ones(polpts)[:],
                                 core.B_tot[-1][:, None, None, None])[-1]

        psi1 = np.broadcast_arrays(np.ones(polpts)[:, None, None, None],
                                   np.ones(inp.numcos)[:, None, None],
                                   np.ones(radpts)[:, None],
                                   np.ones(polpts)[:],
                                   core.psi[-1][:, None, None, None])[-1]

        phi1 = np.broadcast_arrays(np.ones(polpts)[:, None, None, None],
                                   np.ones(inp.numcos)[:, None, None],
                                   np.ones(radpts)[:, None],
                                   np.ones(polpts)[:],
                                   core.E_pot[-1][:, None, None, None])[-1]  # * 1E3  # now in volts

        Tprofile = namedtuple('Tprofile', 'i C')(
            core.T.i.kev.T[0],
            core.T.C.kev.T[0]
        )

        # iol_params = {}
        # iol_params['r0'] = r0
        # iol_params['B0'] = B0
        # iol_params['f0'] = f0
        # iol_params['psi0'] = psi0
        # iol_params['phi0'] = phi0
        # iol_params['zeta0'] = zeta0
        # iol_params['R1'] = R1
        # iol_params['f1'] = f1
        # iol_params['B1'] = B1
        # iol_params['psi1'] = psi1
        # iol_params['phi1'] = phi1
        # iol_params['Tprofile'] = Tprofile

        # convert iol_params to namedtuple so individual parameters can be accessed as normal attributes
        self.iol_p = namedtuple('iol_p', 'r0 B0 f0 psi0 phi0 zeta0 R1 f1 B1 psi1 phi1')(
            r0,
            B0,
            f0,
            psi0,
            phi0,
            zeta0,
            R1,
            f1,
            B1,
            psi1,
            phi1
        )

        # Calculate IOL for thermal deuterium
        forb_d_therm, morb_d_therm, eorb_d_therm = calc_iol_maxwellian(1,
                                                                       m_d,
                                                                       self.iol_p,
                                                                       core.thetapts,
                                                                       Tprofile.i,
                                                                       self.coslist,
                                                                       numcos)
        self.forb_d_therm = inp.R_loss * forb_d_therm
        self.morb_d_therm = inp.R_loss * morb_d_therm
        self.eorb_d_therm = inp.R_loss * eorb_d_therm
        self.forb_d_therm_1D = self.forb_d_therm[:, 0]
        self.morb_d_therm_1D = self.morb_d_therm[:, 0]
        self.eorb_d_therm_1D = self.eorb_d_therm[:, 0]

        # Calculate IOL for thermal tritium
        forb_t_therm, morb_t_therm, eorb_t_therm = calc_iol_maxwellian(1,
                                                                       m_t,
                                                                       self.iol_p,
                                                                       core.thetapts,
                                                                       Tprofile.i,
                                                                       self.coslist,
                                                                       numcos)
        self.forb_t_therm = inp.R_loss * forb_t_therm
        self.morb_t_therm = inp.R_loss * morb_t_therm
        self.eorb_t_therm = inp.R_loss * eorb_t_therm
        self.forb_t_therm_1D = self.forb_t_therm[:, 0]
        self.morb_t_therm_1D = self.morb_t_therm[:, 0]
        self.eorb_t_therm_1D = self.eorb_t_therm[:, 0]

        # Calculate IOL for thermal carbon
        forb_c_therm, morb_c_therm, eorb_c_therm = calc_iol_maxwellian(6,
                                                                       m_c,
                                                                       self.iol_p,
                                                                       core.thetapts,
                                                                       Tprofile.C,
                                                                       self.coslist,
                                                                       numcos)
        self.forb_c_therm = inp.R_loss * forb_c_therm
        self.morb_c_therm = inp.R_loss * morb_c_therm
        self.eorb_c_therm = inp.R_loss * eorb_c_therm
        self.forb_c_therm_1D = self.forb_c_therm[:, 0]
        self.morb_c_therm_1D = self.morb_c_therm[:, 0]
        self.eorb_c_therm_1D = self.eorb_c_therm[:, 0]

        # Calculate IOL for thermal alphas
        forb_a_therm, morb_a_therm, eorb_a_therm = calc_iol_maxwellian(2,
                                                                       m_a,
                                                                       self.iol_p,
                                                                       core.thetapts,
                                                                       Tprofile.i,
                                                                       self.coslist,
                                                                       numcos)
        self.forb_a_therm = inp.R_loss * forb_a_therm
        self.morb_a_therm = inp.R_loss * morb_a_therm
        self.eorb_a_therm = inp.R_loss * eorb_a_therm
        self.forb_a_therm_1D = self.forb_a_therm[:, 0]
        self.morb_a_therm_1D = self.morb_a_therm[:, 0]
        self.eorb_a_therm_1D = self.eorb_a_therm[:, 0]

        # Calculate IOL for fast, monoenergetic alphas
        v_alpha = sqrt(2 * 3.5E6 * 1.6021E-19 / m_a)
        forb_a_fast, morb_a_fast, eorb_a_fast = calc_iol_mono_en(2,
                                                                 m_a,
                                                                 self.iol_p,
                                                                 core.thetapts,
                                                                 v_alpha,
                                                                 self.coslist,
                                                                 numcos)

        # currently applying R_loss to fast alphas as well as thermal, although I'm skeptical of this. -MH
        self.forb_a_fast = inp.R_loss * forb_a_fast
        self.morb_a_fast = inp.R_loss * morb_a_fast
        self.eorb_a_fast = inp.R_loss * eorb_a_fast
        self.forb_a_fast_1D = self.forb_a_fast[:, 0]
        self.morb_a_fast_1D = self.morb_a_fast[:, 0]
        self.eorb_a_fast_1D = self.eorb_a_fast[:, 0]

        # Calculate IOL for neutral deuterium beams
        v_beam = sqrt(2 * 80.0E3 * 1.6021E-19 / m_d)
        zeta_beam = -0.96
        forb_d_nbi, morb_d_nbi, eorb_d_nbi = self.calc_iol_beams(1,
                                                                 m_d,
                                                                 self.iol_p,
                                                                 core.thetapts,
                                                                 v_beam,
                                                                 zeta_beam,
                                                                 self.coslist)
        # currently applying R_loss to fast alphas as well as thermal, although I'm skeptical of this. -MH
        self.forb_d_nbi = inp.R_loss * forb_d_nbi
        self.morb_d_nbi = inp.R_loss * morb_d_nbi
        self.eorb_d_nbi = inp.R_loss * eorb_d_nbi
        self.forb_d_nbi_1D = self.forb_d_nbi[:, 0]
        self.morb_d_nbi_1D = self.morb_d_nbi[:, 0]
        self.eorb_d_nbi_1D = self.eorb_d_nbi[:, 0]

    def plot_F_i(self):
        """
        Plots the 1D GT3.IOL differential ion number loss fraction
        """

        plot = plt.figure()
        fig1 = plot.add_subplot(111)
        fig1.set_xlabel(r'$\rho$', fontsize=20)
        fig1.set_ylabel(r'$F(\rho)$', fontsize=25)
        fig1.set_title('GT3.IOL differential ion number loss fraction')
        fig1.scatter(self.rho, self.forb_d_therm_1D, marker='o', color='blue')

    def plot_M_i(self):
        """
        Plots the 1D GT3.IOL differential ion momentum loss fraction
        """

        plot = plt.figure()
        fig1 = plot.add_subplot(111)
        fig1.set_xlabel(r'$\rho$', fontsize=20)
        fig1.set_ylabel(r'$\M(\rho)$', fontsize=25)
        fig1.set_title('GT3.IOL differential ion momentum loss fraction')
        fig1.scatter(self.rho, self.morb_d_therm_1D, marker='o', color='green')

    def plot_E_i(self):
        """
        Plots the 1D GT3.IOL differential ion energy loss fraction
        """

        plot = plt.figure()
        fig1 = plot.add_subplot(111)
        fig1.set_xlabel(r'$\rho$', fontsize=20)
        fig1.set_ylabel(r'$\frac{\partial E}{\partial r}$', fontsize=25)
        fig1.set_title('GT3.IOL differential ion energy loss fraction')
        fig1.scatter(self.rho, self.eorb_d_therm_1D, marker='o', color='red')

    def plot_all_i(self):
        plot = plt.figure()
        fig1 = plot.add_subplot(111)
        fig1.set_xlabel(r'$\rho$', fontsize=20)
        fig1.set_ylabel(r'$\frac{\partial F}{\partial r}$', fontsize=25)
        fig1.set_title('Ion number loss fraction')
        fig1.scatter(self.rho, self.forb_d_therm_1D, marker='o', color='blue')

        fig2 = plot.add_subplot(121)
        fig2.set_xlabel(r'$\rho$', fontsize=20)
        fig2.set_ylabel(r'$\frac{\partial E}{\partial r}$', fontsize=25)
        fig2.set_title('Ion energy loss fraction')
        fig2.scatter(self.rho, self.eorb_d_therm_1D, marker='o', color='red')

        fig3 = plot.add_subplot(131)
        fig3.set_xlabel(r'$\rho$', fontsize=20)
        fig3.set_ylabel(r'$\frac{\partial M}{\partial r}$', fontsize=25)
        fig3.set_title('Ion momentum loss fraction')
        fig3.scatter(self.rho, self.morb_d_therm_1D, marker='o', color='green')

    def plot_F_C(self):
        """
        Plots the 1D GT3.IOL differential carbon number loss fraction
        """

        plot = plt.figure()
        fig1 = plot.add_subplot(111)
        fig1.set_xlabel(r'$\rho$', fontsize=20)
        fig1.set_ylabel(r'$\frac{\partial F}{\partial r}$', fontsize=25)
        fig1.set_title('GT3.IOL differential carbon number loss fraction')
        fig1.scatter(self.rho, self.forb_c_therm_1D, marker='o', color='blue')

    def plot_M_C(self):
        """
        Plots the 1D GT3.IOL differential carbon momentum loss fraction
        """

        plot = plt.figure()
        fig1 = plot.add_subplot(111)
        fig1.set_xlabel(r'$\rho$', fontsize=20)
        fig1.set_ylabel(r'$\frac{\partial M}{\partial r}$', fontsize=25)
        fig1.set_title('GT3.IOL differential carbon momentum loss fraction')
        fig1.scatter(self.rho, self.morb_c_therm_1D, marker='o', color='green')

    def plot_E_C(self):
        """
        Plots the 1D GT3.IOL differential carbon energy loss fraction
        """

        plot = plt.figure()
        fig1 = plot.add_subplot(111)
        fig1.set_xlabel(r'$\rho$', fontsize=20)
        fig1.set_ylabel(r'$\frac{\partial E}{\partial r}$', fontsize=25)
        fig1.set_title('GT3.IOL differential carbon energy loss fraction')
        fig1.scatter(self.rho, self.eorb_c_therm_1D, marker='o', color='red')

    def plot_all_C(self):
        plot = plt.figure()
        fig1 = plot.add_subplot(111)
        fig1.set_xlabel(r'$\rho$', fontsize=20)
        fig1.set_ylabel(r'$\frac{\partial F}{\partial r}$', fontsize=25)
        fig1.set_title('Carbon number loss fraction')
        fig1.scatter(self.rho, self.forb_c_therm_1D, marker='o', color='blue')

        fig2 = plot.add_subplot(121)
        fig2.set_xlabel(r'$\rho$', fontsize=20)
        fig2.set_ylabel(r'$\frac{\partial E}{\partial r}$', fontsize=25)
        fig2.set_title('Carbon energy loss fraction')
        fig2.scatter(self.rho, self.eorb_c_therm_1D, marker='o', color='red')

        fig3 = plot.add_subplot(131)
        fig3.set_xlabel(r'$\rho$', fontsize=20)
        fig3.set_ylabel(r'$\frac{\partial M}{\partial r}$', fontsize=25)
        fig3.set_title('Carbon momentum loss fraction')
        fig3.scatter(self.rho, self.morb_c_therm_1D, marker='o', color='green')

    def plot_F_i_fast(self):
        """
        Plots the 1D GT3.IOL differential fast ion number loss fraction
        """

        plot = plt.figure()
        fig1 = plot.add_subplot(111)
        fig1.set_xlabel(r'$\rho$', fontsize=20)
        fig1.set_ylabel(r'$\frac{\partial F}{\partial r}$', fontsize=25)
        fig1.set_title('GT3.IOL differential fast ion number loss fraction')
        fig1.scatter(self.rho, self.forb_d_nbi_1D, marker='o', color='blue')

    def plot_M_i_fast(self):
        """
        Plots the 1D GT3.IOL differential fast ion momentum loss fraction
        """

        plot = plt.figure()
        fig1 = plot.add_subplot(111)
        fig1.set_xlabel(r'$\rho$', fontsize=20)
        fig1.set_ylabel(r'$\frac{\partial M}{\partial r}$', fontsize=25)
        fig1.set_title('GT3.IOL differential fast ion momentum loss fraction')
        fig1.scatter(self.rho, self.morb_d_nbi_1D, marker='o', color='green')

    def plot_E_i_fast(self):
        """
        Plots the 1D GT3.IOL differential fast ion energy loss fraction
        """

        plot = plt.figure()
        fig1 = plot.add_subplot(111)
        fig1.set_xlabel(r'$\rho$', fontsize=20)
        fig1.set_ylabel(r'$\frac{\partial E}{\partial r}$', fontsize=25)
        fig1.set_title('GT3.IOL differential fast ion energy loss fraction')
        fig1.scatter(self.rho, self.eorb_d_nbi_1D, marker='o', color='red')

    def plot_all_i_fast(self):
        plot = plt.figure()
        fig1 = plot.add_subplot(111)
        fig1.set_xlabel(r'$\rho$', fontsize=20)
        fig1.set_ylabel(r'$\frac{\partial F}{\partial r}$', fontsize=25)
        fig1.set_title('Fast ion number loss fraction')
        fig1.scatter(self.rho, self.forb_d_nbi_1D, marker='o', color='blue')

        fig2 = plot.add_subplot(121)
        fig2.set_xlabel(r'$\rho$', fontsize=20)
        fig2.set_ylabel(r'$\frac{\partial E}{\partial r}$', fontsize=25)
        fig2.set_title('Fast ion energy loss fraction')
        fig2.scatter(self.rho, self.eorb_d_nbi_1D, marker='o', color='red')

        fig3 = plot.add_subplot(131)
        fig3.set_xlabel(r'$\rho$', fontsize=20)
        fig3.set_ylabel(r'$\frac{\partial M}{\partial r}$', fontsize=25)
        fig3.set_title('Fast ion momentum loss fraction')
        fig3.scatter(self.rho, self.morb_d_nbi_1D, marker='o', color='green')

# iolplot=1
# if iolplot==1:
#    fig = plt.figure(figsize=(5, 5))
#    fig.suptitle('IOL a, b, c in DIII-D with cos:{}'.format(coslist[-2]), fontsize=15)
#    ax1 = fig.add_subplot(221)
#    ax1.set_title(r'$v_{sep-1}$', fontsize=12)
#    ax1.set_ylabel(r'R', fontsize=12)
#    ax1.set_xlabel(r'Z', fontsize=12)
#    ax1.axis('equal')
#    ax1.grid(b=True, which='both', axis='both')
#    CS = ax1.contourf(brnd.R, brnd.Z, v_sep_1[:, -2, :, 0].T, 500)
#    plt.colorbar(CS)
#
#    ax2 = fig.add_subplot(222)
#    ax2.set_title(r'$v_{sep-2}$', fontsize=12)
#    ax2.set_ylabel(r'R', fontsize=12)
#    ax2.set_xlabel(r'Z', fontsize=12)
#    ax2.axis('equal')
#    ax2.grid(b=True, which='both', axis='both')
#    CS = ax2.contourf(brnd.R, brnd.Z, v_sep_2[:, -2, :, 0].T, 500)
#    plt.colorbar(CS)
#
#    ax3 = fig.add_subplot(223)
#    ax3.set_title(r'$v_{sep}$', fontsize=12)
#    ax3.set_ylabel(r'R', fontsize=12)
#    ax3.set_xlabel(r'Z', fontsize=12)
#    ax3.axis('equal')
#    ax3.grid(b=True, which='both', axis='both')
#    #CS = ax3.pcolor(brnd.R, brnd.Z, v_sep[:, 10, :, 0].T, vmin=0, vmax=1E8)
#    CS = ax3.contourf(brnd.R, brnd.Z, v_sep[:, -2, :, 0].T, 500)
#    plt.colorbar(CS)

#    plt.tight_layout()
#    fig.subplots_adjust(top=0.84)
# if iolplot==1:
#     fig = plt.figure(figsize=(6, 8))
#     fig.suptitle(r'IOL v_{} with cos:{}'.format('esc', coslist[0]), fontsize=15)
#     ax1 = fig.add_subplot(111)
#     #ax1.set_title(r'$v_{sep}$', fontsize=12)
#     ax1.set_ylabel(r'Z', fontsize=12)
#     ax1.set_xlabel(r'R', fontsize=12)
#     ax1.axis('equal')
#     ax1.grid(b=True, which='both', axis='both')
#
#     test = np.nanmin(v_sep, axis=3).T
#
#     CS = ax1.contourf(brnd.R, brnd.Z, np.log10((0.5*m*test[:, 0, :]**2)/(1.6021E-19 * 1.0E3)), 500)
#     plt.colorbar(CS)
#     #plt.tight_layout()
#     fig.subplots_adjust(top=0.95)
#     #fig.subplots_adjust(top=0.95)
# return



#!/usr/bin/python


from numpy import log as ln
from scipy.interpolate import UnivariateSpline
import numpy as np

def get_mfp(ne, beamE, Te, beamA, Zeff, rho):
    """
    Returns the mean free path as a Univariate Spline

    :return: A univariate spline of the MFP function
    :param x: The rho value (in [0., 1]
    :type x: float
    :param ne: The elecron density ndarray
    :type ne:  np.ndarray
    :param beamE: The beam energy
    :type beamE:  float
    :param Te: The electron temperature array
    :type Te:  np.ndarray
    :param beamA: The beam ion mass in u
    :type beamA:  float
    :param Zeff:  The Z-effective array
    :type Zeff:  np.ndarray
    :return:
    :param rho: The rho array
    :type rho:  np.ndarray

    :return:
    """

    ne_spline = UnivariateSpline(rho, ne / 1E6)
    sig_eff_spline = sigeff(ne, beamE, Te, beamA, Zeff, rho)
    result = np.array([(1. / (ne_spline(x) * sig_eff_spline(x))) * 1E-2 for x in rho])

    if not (result > 0).all():
        raise ValueError.message("Negative MFP encountered")
    else:
        return UnivariateSpline(rho, result, k=3, s=2)

def sigeff(ne, beamE, Te, beamA, Zeff, rho):

    """
    Calculates effective cross sections for a single-impurity plasma. Returns a Univariate spline.

    Reference:
    R.K. Janev, C.D. Boley, D.E. Post. Penetration of energetic neutral beams into fusion plasmas.
    NUCLEAR  FUSION,  Vol.29,  No. 12  (1989)

    :param kwargs:
    :return: A Univariate spline
    :param rho: The rho array
    :type rho: np.array
    :param E: The beam energy
    :type float
    :param ne: The electron density
    :type ne: np.array
    :param Te: The electron temperature
    :type Te: np.array
    :param Zeff: The effective Z
    :type Zeff: np.array
    :param nz: The impurity density
    :type nz: np.array
    """


    expTerm = np.exp(S1(ne, beamE, Te, beamA, rho)) / (beamE / beamA)
    rest = (1 + (Zeff - 1) * SC(ne, beamE, Te, beamA)) * 1E-16
    fun = expTerm * rest
    return UnivariateSpline(rho, fun)


def S1(ne, E, Te, beamA, rho):
    """
    Calculates S1 coefficient for hydrogenic plasma impurities

    :param ne: Electron density (in m^-3)
    :type ne: np.ndarray
    :param E: Beam ion temperature  (in eV)
    :type E: np.ndarray
    :param Te: Plasma electron temperature  (in keV)
    :type Te: np.ndarray
    :param beamA: Atomic mass of beam ions (in u)
    :type beamA: float
    :return result: The rho-dependent S1
    :type result: np.array
    """

    A1 = np.zeros((3, 3, 2))
    A1[0, 0, 0] = 3.95E+00
    A1[0, 0, 1] = 1.60E-02
    A1[0, 1, 0] = -3.84E-02
    A1[0, 1, 1] = -5.98E-03
    A1[0, 2, 0] = -3.10E-03
    A1[0, 2, 1] = -1.09E-03
    A1[1, 0, 0] = 3.67E-01
    A1[1, 0, 1] = -2.15E-02
    A1[1, 1, 0] = 3.07E-02
    A1[1, 1, 1] = 1.78E-03
    A1[1, 2, 0] = 3.16E-03
    A1[1, 2, 1] = 3.47E-04
    A1[2, 0, 0] = -9.95E-03
    A1[2, 0, 1] = 6.19E-04
    A1[2, 1, 0] = -2.36E-03
    A1[2, 1, 1] = -1.67E-04
    A1[2, 2, 0] = -1.31E-04
    A1[2, 2, 1] = -2.28E-05

    result = np.zeros(len(rho))
    for n, val in enumerate(rho):
        for i in [0,1]:
            for j in [0, 1, 2]:
                for k in [0, 1]:
                    result[n] += A1[i, j, k] * (ln(E * 1.E-3/beamA)**(i)) * ( ( ln((ne[n] / 1E6) / 1E13) ) )**(j) * (ln(Te[n]))**(k)
    return result

def SC(ne, E, Te, beamA):
    """
    Calculates S2 coefficient for Carbon impurities

    :param ne: Electron density (in m^-3)
    :type ne: np.ndarray
    :param E: Beam ion temperature  (in eV)
    :type E: np.ndarray
    :param Te: Electron temperature  (in keV)
    :type Te: np.ndarray
    :param beamA: Atomic mass of beam ions (in u)
    :type beamA: float
    :return result: The rho-dependent SC
    :type result: np.array
    :return:
    """
    # TODO: Get improved coefficients since Mandrekas used improved coefficients in S1

    BC = np.zeros((4, 2, 2))
    BC[0, 0, 0] = -1.49E0
    BC[0, 0, 1] = -1.54E-2
    BC[0, 1, 0] = -1.19E-2
    BC[0, 1, 1] = -1.50E-2
    BC[1, 0, 0] = 5.18E-1
    BC[1, 0, 1] = 7.18E-3
    BC[1, 1, 0] = 2.92E-2
    BC[1, 1, 1] = 3.66E-3
    BC[2, 0, 0] = -3.36E-2
    BC[2, 0, 1] = 3.41E-4
    BC[2, 1, 0] = -1.79E-3
    BC[2, 1, 1] = -2.14E-4


    result = np.zeros(len(ne))
    for rho in range(0, len(ne)):
        for i in [0, 1, 2]:
            for j in [0, 1]:
                for k in [0, 1]:
                    result[rho] += BC[i, j, k] * (ln(E * 1.E-3/beamA)**(i)) * (ln((ne[rho] / 1E6) / 1E13))**(j)\
                                   * (ln(Te[rho]))**(k)
    return result


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

from collections import namedtuple

def combine_objs(obj1, obj2):
    """Combines dictionaries of two instantiated classes or namedtuples into a new namedtuple"""

    #this function might need some work, just FYI.
    try:
        dict1 = obj1._asdict()
    except AttributeError:
        dict1 = obj1.__dict__

    try:
        dict2 = obj2._asdict()
    except AttributeError:
        dict2 = obj2.__dict__

    temp = dict2.copy()
    new_dict = dict1.update(dict2)
    new_nt = namedtuple('new', new_dict.keys())(*new_dict.values())
    return new_nt
